"""
The crypt module manages all of the cryptography functions for minions and
masters, encrypting and decrypting payloads, preparing messages, and
authenticating peers
"""

import base64
import binascii
import copy
import getpass
import hashlib
import hmac
import logging
import os
import pathlib
import random
import stat
import sys
import tempfile
import time
import traceback
import uuid
import weakref

import salt.channel.client
import salt.defaults.exitcodes
import salt.ext.tornado.gen
import salt.payload
import salt.utils.crypt
import salt.utils.decorators
import salt.utils.event
import salt.utils.files
import salt.utils.rsax931
import salt.utils.sdb
import salt.utils.stringutils
import salt.utils.user
import salt.utils.verify
import salt.version
from salt.exceptions import (
    AuthenticationError,
    InvalidKeyError,
    MasterExit,
    SaltClientError,
    SaltReqTimeoutError,
    UnsupportedAlgorithm,
)

try:
    import cryptography.exceptions
    from cryptography.hazmat.primitives import hashes, serialization
    from cryptography.hazmat.primitives.asymmetric import padding, rsa
    from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

    HAS_CRYPTOGRAPHY = True
except ImportError:
    HAS_CRYPTOGRAPHY = False


log = logging.getLogger(__name__)

OAEP = "OAEP"
PKCS1v15 = "PKCS1v15"

SHA1 = "SHA1"
SHA224 = "SHA224"

OAEP_SHA1 = f"{OAEP}-{SHA1}"
OAEP_SHA224 = f"{OAEP}-{SHA224}"

PKCS1v15_SHA1 = f"{PKCS1v15}-{SHA1}"
PKCS1v15_SHA224 = f"{PKCS1v15}-{SHA224}"


VALID_HASHES = (
    SHA1,
    SHA224,
)

VALID_PADDING_FOR_SIGNING = (PKCS1v15,)
VALID_PADDING_FOR_ENCRYPTION = (OAEP,)
VALID_ENCRYPTION_ALGORITHMS = (
    OAEP_SHA1,
    OAEP_SHA224,
)
VALID_SIGNING_ALGORITHMS = (
    PKCS1v15_SHA1,
    PKCS1v15_SHA224,
)


def fips_enabled():
    if HAS_CRYPTOGRAPHY:
        import cryptography.hazmat.backends.openssl.backend

        return cryptography.hazmat.backends.openssl.backend._fips_enabled


def clean_key(key):
    """
    Clean the key so that it only has unix style line endings (\\n)
    """
    return "\n".join(key.strip().splitlines())


def dropfile(cachedir, user=None):
    """
    Set an AES dropfile to request the master update the publish session key
    """
    dfn = os.path.join(cachedir, ".dfn")
    # set a mask (to avoid a race condition on file creation) and store original.
    with salt.utils.files.set_umask(0o277):
        log.info("Rotating AES key")
        if os.path.isfile(dfn):
            log.info("AES key rotation already requested")
            return

        if os.path.isfile(dfn) and not os.access(dfn, os.W_OK):
            os.chmod(dfn, stat.S_IRUSR | stat.S_IWUSR)
        with salt.utils.files.fopen(dfn, "wb+") as fp_:
            fp_.write(b"")
        os.chmod(dfn, stat.S_IRUSR)
        if user:
            try:
                import pwd

                uid = pwd.getpwnam(user).pw_uid
                os.chown(dfn, uid, -1)
            except (KeyError, ImportError, OSError):
                pass


def gen_keys(keydir, keyname, keysize, user=None, passphrase=None, e=65537):
    """
    Generate a RSA public keypair for use with salt

    :param str keydir: The directory to write the keypair to
    :param str keyname: The type of salt server for whom this key should be written. (i.e. 'master' or 'minion')
    :param int keysize: The number of bits in the key
    :param str user: The user on the system who should own this keypair
    :param str passphrase: The passphrase which should be used to encrypt the private key

    :rtype: str
    :return: Path on the filesystem to the RSA private key
    """
    base = os.path.join(keydir, keyname)
    priv = f"{base}.pem"
    pub = f"{base}.pub"

    gen = rsa.generate_private_key(e, keysize)

    if os.path.isfile(priv):
        # Between first checking and the generation another process has made
        # a key! Use the winner's key
        return priv

    # Do not try writing anything, if directory has no permissions.
    if not os.access(keydir, os.W_OK):
        raise OSError(
            'Write access denied to "{}" for user "{}".'.format(
                os.path.abspath(keydir), getpass.getuser()
            )
        )

    with salt.utils.files.set_umask(0o277):
        with salt.utils.files.fopen(priv, "wb+") as f:
            if passphrase:
                enc = serialization.BestAvailableEncryption(passphrase.encode())
                _format = serialization.PrivateFormat.TraditionalOpenSSL
                if fips_enabled():
                    _format = serialization.PrivateFormat.PKCS8
            else:
                enc = serialization.NoEncryption()
                _format = serialization.PrivateFormat.TraditionalOpenSSL
            pem = gen.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=_format,
                encryption_algorithm=enc,
            )
            f.write(pem)

    pubkey = gen.public_key()
    with salt.utils.files.fopen(pub, "wb+") as f:
        pem = pubkey.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )
        f.write(pem)

    os.chmod(priv, 0o400)
    if user:
        try:
            import pwd

            uid = pwd.getpwnam(user).pw_uid
            os.chown(priv, uid, -1)
            os.chown(pub, uid, -1)
        except (KeyError, ImportError, OSError):
            # The specified user was not found, allow the backup systems to
            # report the error
            pass
    return priv


class BaseKey:

    @staticmethod
    def parse_padding_for_signing(algorithm):
        if algorithm not in VALID_SIGNING_ALGORITHMS:
            raise UnsupportedAlgorithm(f"Invalid signing algorithm: {algorithm}")
        _pad, _hash = algorithm.split("-", 1)
        if _pad not in VALID_PADDING_FOR_SIGNING:
            raise UnsupportedAlgorithm(f"Invalid padding algorithm: {_pad}")
        return getattr(padding, _pad)

    @staticmethod
    def parse_padding_for_encryption(algorithm):
        if algorithm not in VALID_ENCRYPTION_ALGORITHMS:
            raise UnsupportedAlgorithm(f"Invalid encryption algorithm: {algorithm}")
        _pad, _hash = algorithm.split("-", 1)
        if _pad not in VALID_PADDING_FOR_ENCRYPTION:
            raise UnsupportedAlgorithm(f"Invalid padding algorithm: {_pad}")
        return getattr(padding, _pad)

    @staticmethod
    def parse_hash(algorithm):
        if "-" not in algorithm:
            raise UnsupportedAlgorithm(f"Invalid encryption algorithm: {algorithm}")
        _pad, _hash = algorithm.split("-", 1)
        if _hash not in VALID_HASHES:
            raise Exception("Invalid hashing algorithm")
        return getattr(hashes, _hash)


class PrivateKey(BaseKey):

    def __init__(self, path, passphrase=None):
        self.key = get_rsa_key(path, passphrase)

    def encrypt(self, data):
        pem = self.key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption(),
        )
        return salt.utils.rsax931.RSAX931Signer(pem).sign(data)

    def sign(self, data, algorithm=PKCS1v15_SHA1):
        _padding = self.parse_padding_for_signing(algorithm)
        _hash = self.parse_hash(algorithm)
        try:
            return self.key.sign(
                salt.utils.stringutils.to_bytes(data), _padding(), _hash()
            )
        except cryptography.exceptions.UnsupportedAlgorithm:
            raise UnsupportedAlgorithm(f"Unsupported algorithm: {algorithm}")

    def decrypt(self, data, algorithm=OAEP_SHA1):
        _padding = self.parse_padding_for_encryption(algorithm)
        _hash = self.parse_hash(algorithm)
        try:
            return self.key.decrypt(
                data,
                _padding(
                    mgf=padding.MGF1(algorithm=_hash()),
                    algorithm=_hash(),
                    label=None,
                ),
            )
        except cryptography.exceptions.UnsupportedAlgorithm:
            raise UnsupportedAlgorithm(f"Unsupported algorithm: {algorithm}")


class PublicKey(BaseKey):
    def __init__(self, path):
        with salt.utils.files.fopen(path, "rb") as fp:
            try:
                self.key = serialization.load_pem_public_key(fp.read())
            except ValueError as exc:
                raise InvalidKeyError("Invalid key")

    def encrypt(self, data, algorithm=OAEP_SHA1):
        _padding = self.parse_padding_for_encryption(algorithm)
        _hash = self.parse_hash(algorithm)
        bdata = salt.utils.stringutils.to_bytes(data)
        try:
            return self.key.encrypt(
                bdata,
                _padding(
                    mgf=padding.MGF1(algorithm=_hash()),
                    algorithm=_hash(),
                    label=None,
                ),
            )
        except cryptography.exceptions.UnsupportedAlgorithm:
            raise UnsupportedAlgorithm(f"Unsupported algorithm: {algorithm}")

    def verify(self, data, signature, algorithm=PKCS1v15_SHA1):
        _padding = self.parse_padding_for_signing(algorithm)
        _hash = self.parse_hash(algorithm)
        try:
            self.key.verify(
                salt.utils.stringutils.to_bytes(signature),
                salt.utils.stringutils.to_bytes(data),
                _padding(),
                _hash(),
            )
        except cryptography.exceptions.InvalidSignature:
            return False
        return True

    def decrypt(self, data):
        pem = self.key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )
        verifier = salt.utils.rsax931.RSAX931Verifier(pem)
        return verifier.verify(data)


@salt.utils.decorators.memoize
def _get_key_with_evict(path, timestamp, passphrase):
    """
    Load a private key from disk.  `timestamp` above is intended to be the
    timestamp of the file's last modification. This fn is memoized so if it is
    called with the same path and timestamp (the file's last modified time) the
    second time the result is returned from the memoization.  If the file gets
    modified then the params are different and the key is loaded from disk.
    """
    log.debug("salt.crypt._get_key_with_evict: Loading private key")
    if passphrase:
        password = passphrase.encode()
    else:
        password = None
    with salt.utils.files.fopen(path, "rb") as f:
        try:
            return serialization.load_pem_private_key(
                f.read(),
                password=password,
            )
        except ValueError:
            raise InvalidKeyError("Encountered bad RSA public key")
        except cryptography.exceptions.UnsupportedAlgorithm:
            raise InvalidKeyError("Unsupported key algorithm")


def get_rsa_key(path, passphrase):
    """
    Read a private key off the disk.  Poor man's simple cache in effect here,
    we memoize the result of calling _get_rsa_with_evict.  This means the first
    time _get_key_with_evict is called with a path and a timestamp the result
    is cached.  If the file (the private key) does not change then its
    timestamp will not change and the next time the result is returned from the
    cache.  If the key DOES change the next time _get_rsa_with_evict is called
    it is called with different parameters and the fn is run fully to retrieve
    the key from disk.
    """
    log.debug("salt.crypt.get_rsa_key: Loading private key")
    return _get_key_with_evict(path, str(os.path.getmtime(path)), passphrase)


def get_rsa_pub_key(path):
    """
    Read a public key off the disk.
    """
    log.debug("salt.crypt.get_rsa_pub_key: Loading public key")
    try:
        with salt.utils.files.fopen(path, "rb") as fp:
            return serialization.load_pem_public_key(fp.read())
    except ValueError:
        raise InvalidKeyError("Encountered bad RSA public key")
    except cryptography.exceptions.UnsupportedAlgorithm:
        raise InvalidKeyError("Unsupported key algorithm")


def sign_message(privkey_path, message, passphrase=None, algorithm=PKCS1v15_SHA1):
    """
    Use Crypto.Signature.PKCS1_v1_5 to sign a message. Returns the signature.
    """
    return PrivateKey(privkey_path, passphrase).sign(message, algorithm)


def verify_signature(pubkey_path, message, signature, algorithm=PKCS1v15_SHA1):
    """
    Use Crypto.Signature.PKCS1_v1_5 to verify the signature on a message.
    Returns True for valid signature.
    """
    log.debug("salt.crypt.verify_signature: Loading public key")
    return PublicKey(pubkey_path).verify(message, signature, algorithm)


def gen_signature(priv_path, pub_path, sign_path, passphrase=None):
    """
    creates a signature for the given public-key with
    the given private key and writes it to sign_path
    """

    with salt.utils.files.fopen(pub_path) as fp_:
        mpub_64 = fp_.read()

    mpub_sig = sign_message(priv_path, mpub_64, passphrase)
    mpub_sig_64 = binascii.b2a_base64(mpub_sig)
    if os.path.isfile(sign_path):
        return False
    log.trace(
        "Calculating signature for %s with %s",
        os.path.basename(pub_path),
        os.path.basename(priv_path),
    )

    if os.path.isfile(sign_path):
        log.trace(
            "Signature file %s already exists, please remove it first and try again",
            sign_path,
        )
    else:
        with salt.utils.files.fopen(sign_path, "wb+") as sig_f:
            sig_f.write(salt.utils.stringutils.to_bytes(mpub_sig_64))
        log.trace("Wrote signature to %s", sign_path)
    return True


def private_encrypt(key, message):
    """
    Generate an M2Crypto-compatible signature

    :param Crypto.PublicKey.RSA._RSAobj key: The RSA key object
    :param str message: The message to sign
    :rtype: str
    :return: The signature, or an empty string if the signature operation failed
    """
    return key.encrypt(message)


def pwdata_decrypt(rsa_key, pwdata):
    key = serialization.load_pem_private_key(rsa_key.encode(), password=None)
    password = key.decrypt(
        pwdata,
        padding.PKCS1v15(),
    )
    return salt.utils.stringutils.to_unicode(password)


class MasterKeys(dict):
    """
    The Master Keys class is used to manage the RSA public key pair used for
    authentication by the master.

    It also generates a signing key-pair if enabled with master_sign_key_name.
    """

    def __init__(self, opts):
        super().__init__()
        self.opts = opts
        self.pub_path = os.path.join(self.opts["pki_dir"], "master.pub")
        self.rsa_path = os.path.join(self.opts["pki_dir"], "master.pem")

        key_pass = salt.utils.sdb.sdb_get(self.opts["key_pass"], self.opts)
        self.key = self.__get_keys(passphrase=key_pass)

        self.pub_signature = None

        # set names for the signing key-pairs
        if opts["master_sign_pubkey"]:

            # if only the signature is available, use that
            if opts["master_use_pubkey_signature"]:
                self.sig_path = os.path.join(
                    self.opts["pki_dir"], opts["master_pubkey_signature"]
                )
                if os.path.isfile(self.sig_path):
                    with salt.utils.files.fopen(self.sig_path) as fp_:
                        self.pub_signature = clean_key(fp_.read())
                    log.info(
                        "Read %s's signature from %s",
                        os.path.basename(self.pub_path),
                        self.opts["master_pubkey_signature"],
                    )
                else:
                    log.error(
                        "Signing the master.pub key with a signature is "
                        "enabled but no signature file found at the defined "
                        "location %s",
                        self.sig_path,
                    )
                    log.error(
                        "The signature-file may be either named differently "
                        "or has to be created with 'salt-key --gen-signature'"
                    )
                    sys.exit(1)

            # create a new signing key-pair to sign the masters
            # auth-replies when a minion tries to connect
            else:
                key_pass = salt.utils.sdb.sdb_get(
                    self.opts["signing_key_pass"], self.opts
                )
                self.pub_sign_path = os.path.join(
                    self.opts["pki_dir"], opts["master_sign_key_name"] + ".pub"
                )
                self.rsa_sign_path = os.path.join(
                    self.opts["pki_dir"], opts["master_sign_key_name"] + ".pem"
                )
                self.sign_key = self.__get_keys(name=opts["master_sign_key_name"])

    # We need __setstate__ and __getstate__ to avoid pickling errors since
    # some of the member variables correspond to Cython objects which are
    # not picklable.
    # These methods are only used when pickling so will not be used on
    # non-Windows platforms.
    def __setstate__(self, state):
        self.__init__(state["opts"])

    def __getstate__(self):
        return {"opts": self.opts}

    def __get_keys(self, name="master", passphrase=None):
        """
        Returns a key object for a key in the pki-dir
        """
        path = os.path.join(self.opts["pki_dir"], name + ".pem")
        if not os.path.exists(path):
            log.info("Generating %s keys: %s", name, self.opts["pki_dir"])
            gen_keys(
                self.opts["pki_dir"],
                name,
                self.opts["keysize"],
                self.opts.get("user"),
                passphrase,
            )
        try:
            key = PrivateKey(path, passphrase)
        except ValueError as e:
            message = f"Unable to read key: {path}; file may be corrupt"
        except TypeError as e:
            message = f"Unable to read key: {path}; passphrase may be incorrect"
        except InvalidKeyError as e:
            message = f"Unable to read key: {path}; key contains unsupported algorithm"
        except cryptography.exceptions.UnsupportedAlgorithm as e:
            message = f"Unable to read key: {path}; key contains unsupported algorithm"
        else:
            log.debug("Loaded %s key: %s", name, path)
            return key
        log.error(message)
        raise MasterExit(message)

    def get_pub_str(self, name="master"):
        """
        Return the string representation of a public key
        in the pki-directory
        """
        path = os.path.join(self.opts["pki_dir"], name + ".pub")
        if not os.path.isfile(path):
            pubkey = self.key.public_key()
            with salt.utils.files.fopen(path, "wb+") as f:
                f.write(
                    pubkey.public_bytes(
                        encoding=serialization.Encoding.PEM,
                        format=serialization.PublicFormat.SubjectPublicKeyInfo,
                    )
                )
        with salt.utils.files.fopen(path) as rfh:
            return clean_key(rfh.read())

    def get_mkey_paths(self):
        return self.pub_path, self.rsa_path

    def get_sign_paths(self):
        return self.pub_sign_path, self.rsa_sign_path

    def pubkey_signature(self):
        """
        returns the base64 encoded signature from the signature file
        or None if the master has its own signing keys
        """
        return self.pub_signature


class AsyncAuth:
    """
    Set up an Async object to maintain authentication with the salt master
    """

    # This class is only a singleton per minion/master pair
    # mapping of io_loop -> {key -> auth}
    instance_map = weakref.WeakKeyDictionary()

    # mapping of key -> creds
    creds_map = {}

    def __new__(cls, opts, io_loop=None):
        """
        Only create one instance of AsyncAuth per __key()
        """
        # do we have any mapping for this io_loop
        io_loop = io_loop or salt.ext.tornado.ioloop.IOLoop.current()
        if io_loop not in AsyncAuth.instance_map:
            AsyncAuth.instance_map[io_loop] = weakref.WeakValueDictionary()
        loop_instance_map = AsyncAuth.instance_map[io_loop]

        key = cls.__key(opts)
        auth = loop_instance_map.get(key)
        if auth is None:
            log.debug("Initializing new AsyncAuth for %s", key)
            # we need to make a local variable for this, as we are going to store
            # it in a WeakValueDictionary-- which will remove the item if no one
            # references it-- this forces a reference while we return to the caller
            auth = object.__new__(cls)
            auth.__singleton_init__(opts, io_loop=io_loop)
            loop_instance_map[key] = auth
        else:
            log.debug("Re-using AsyncAuth for %s", key)
        return auth

    @classmethod
    def __key(cls, opts, io_loop=None):
        keypath = os.path.join(opts["pki_dir"], "minion.pem")
        keytime = "0"
        if os.path.exists(keypath):
            keytime = str(os.path.getmtime(keypath))
        return (
            opts["pki_dir"],  # where the keys are stored
            opts["id"],  # minion ID
            opts["master_uri"],  # master ID
            keytime,
        )

    # has to remain empty for singletons, since __init__ will *always* be called
    def __init__(self, opts, io_loop=None):
        pass

    # an init for the singleton instance to call
    def __singleton_init__(self, opts, io_loop=None):
        """
        Init an Auth instance

        :param dict opts: Options for this server
        :return: Auth instance
        :rtype: Auth
        """
        self.opts = opts
        self.token = salt.utils.stringutils.to_bytes(Crypticle.generate_key_string())
        self.pub_path = os.path.join(self.opts["pki_dir"], "minion.pub")
        self.rsa_path = os.path.join(self.opts["pki_dir"], "minion.pem")
        self._private_key = None
        if self.opts["__role"] == "syndic":
            self.mpub = "syndic_master.pub"
        else:
            self.mpub = "minion_master.pub"
        if not os.path.isfile(self.pub_path):
            self.get_keys()

        self.io_loop = io_loop or salt.ext.tornado.ioloop.IOLoop.current()
        key = self.__key(self.opts)
        # TODO: if we already have creds for this key, lets just re-use
        if key in AsyncAuth.creds_map:
            creds = AsyncAuth.creds_map[key]
            self._creds = creds
            self._crypticle = Crypticle(self.opts, creds["aes"])
            self._session_crypticle = Crypticle(self.opts, creds["session"])
            self._authenticate_future = salt.ext.tornado.concurrent.Future()
            self._authenticate_future.set_result(True)
        else:
            self.authenticate()

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls, copy.deepcopy(self.opts, memo))
        memo[id(self)] = result
        for key in self.__dict__:
            if key in ("io_loop",):
                # The io_loop has a thread Lock which will fail to be deep
                # copied. Skip it because it will just be recreated on the
                # new copy.
                continue
            setattr(result, key, copy.deepcopy(self.__dict__[key], memo))
        return result

    @property
    def session_crypticle(self):
        return self._session_crypticle

    @property
    def creds(self):
        return self._creds

    @property
    def crypticle(self):
        return self._crypticle

    @property
    def authenticated(self):
        return (
            hasattr(self, "_authenticate_future")
            and self._authenticate_future.done()
            and self._authenticate_future.exception() is None
        )

    def invalidate(self):
        if self.authenticated:
            del self._authenticate_future
            key = self.__key(self.opts)
            if key in AsyncAuth.creds_map:
                del AsyncAuth.creds_map[key]

    def authenticate(self, callback=None):
        """
        Ask for this client to reconnect to the origin

        This function will de-dupe all calls here and return a *single* future
        for the sign-in-- whis way callers can all assume there aren't others
        """
        # if an auth is in flight-- and not done-- just pass that back as the future to wait on
        if (
            hasattr(self, "_authenticate_future")
            and not self._authenticate_future.done()
        ):
            future = self._authenticate_future
        else:
            future = salt.ext.tornado.concurrent.Future()
            self._authenticate_future = future
            self.io_loop.add_callback(self._authenticate)

        if callback is not None:

            def handle_future(future):
                response = future.result()
                self.io_loop.add_callback(callback, response)

            future.add_done_callback(handle_future)

        return future

    @salt.ext.tornado.gen.coroutine
    def _authenticate(self):
        """
        Authenticate with the master, this method breaks the functional
        paradigm, it will update the master information from a fresh sign
        in, signing in can occur as often as needed to keep up with the
        revolving master AES key.

        :rtype: Crypticle
        :returns: A crypticle used for encryption operations
        """
        acceptance_wait_time = self.opts["acceptance_wait_time"]
        acceptance_wait_time_max = self.opts["acceptance_wait_time_max"]
        if not acceptance_wait_time_max:
            acceptance_wait_time_max = acceptance_wait_time
        creds = None

        with salt.channel.client.AsyncReqChannel.factory(
            self.opts, crypt="clear", io_loop=self.io_loop
        ) as channel:
            error = None
            while True:
                try:
                    creds = yield self.sign_in(channel=channel)
                except SaltClientError as exc:
                    error = exc
                    break
                if creds == "retry":
                    if self.opts.get("detect_mode") is True:
                        error = SaltClientError("Detect mode is on")
                        break
                    if self.opts.get("caller"):
                        # We have a list of masters, so we should break
                        # and try the next one in the list.
                        if self.opts.get("local_masters", None):
                            error = SaltClientError(
                                "Minion failed to authenticate"
                                " with the master, has the "
                                "minion key been accepted?"
                            )
                            break
                        else:
                            print(
                                "Minion failed to authenticate with the master, "
                                "has the minion key been accepted?"
                            )
                            sys.exit(2)
                    if acceptance_wait_time:
                        log.info(
                            "Waiting %s seconds before retry.", acceptance_wait_time
                        )
                        yield salt.ext.tornado.gen.sleep(acceptance_wait_time)
                    if acceptance_wait_time < acceptance_wait_time_max:
                        acceptance_wait_time += acceptance_wait_time
                        log.debug(
                            "Authentication wait time is %s", acceptance_wait_time
                        )
                    continue
                elif creds == "bad enc algo":
                    log.error(
                        "This minion is using a encryption algorithm that is "
                        "not supported by it's Master. Please check your minion configutation."
                    )
                    break
                elif creds == "bad sig algo":
                    log.error(
                        "This minion is using a signing algorithm that is "
                        "not supported by it's Master. Please check your minion configutation."
                    )
                    break
                break
            if not isinstance(creds, dict) or "aes" not in creds:
                if self.opts.get("detect_mode") is True:
                    error = SaltClientError("-|RETRY|-")
                try:
                    del AsyncAuth.creds_map[self.__key(self.opts)]
                except KeyError:
                    pass
                if not error:
                    error = SaltClientError(
                        "Attempt to authenticate with the salt master failed"
                    )
                self._authenticate_future.set_exception(error)
            else:
                key = self.__key(self.opts)
                new_aes, changed_aes, changed_session = False, False, False
                if key not in AsyncAuth.creds_map:
                    new_aes = True
                    log.debug("%s Got new master aes key.", self)
                else:
                    if self._creds["aes"] != creds["aes"]:
                        changed_aes = True
                        log.debug("%s The master's aes key has changed.", self)
                    if self._creds["session"] != creds["session"]:
                        changed_session = True
                        log.debug("%s The master's session key has changed.", self)
                if new_aes or changed_aes or changed_session:
                    AsyncAuth.creds_map[key] = creds
                    self._creds = creds
                    self._crypticle = Crypticle(self.opts, creds["aes"])
                    self._session_crypticle = Crypticle(self.opts, creds["session"])

                self._authenticate_future.set_result(
                    True
                )  # mark the sign-in as complete
                # Notify the bus about creds change
                if self.opts.get("auth_events") is True:
                    with salt.utils.event.get_event(
                        self.opts.get("__role"), opts=self.opts, listen=False
                    ) as event:
                        event.fire_event(
                            {"key": key, "creds": creds},
                            salt.utils.event.tagify(prefix="auth", suffix="creds"),
                        )

    @salt.ext.tornado.gen.coroutine
    def sign_in(self, timeout=60, safe=True, tries=1, channel=None):
        """
        Send a sign in request to the master, sets the key information and
        returns a dict containing the master publish interface to bind to
        and the decrypted aes key for transport decryption.

        :param int timeout: Number of seconds to wait before timing out the sign-in request
        :param bool safe: If True, do not raise an exception on timeout. Retry instead.
        :param int tries: The number of times to try to authenticate before giving up.

        :raises SaltReqTimeoutError: If the sign-in request has timed out and :param safe: is not set

        :return: Return a string on failure indicating the reason for failure. On success, return a dictionary
        with the publication port and the shared AES key.

        """
        auth_timeout = self.opts.get("auth_timeout", None)
        if auth_timeout is not None:
            timeout = auth_timeout
        auth_safemode = self.opts.get("auth_safemode", None)
        if auth_safemode is not None:
            safe = auth_safemode
        auth_tries = self.opts.get("auth_tries", None)
        if auth_tries is not None:
            tries = auth_tries

        close_channel = False
        if not channel:
            close_channel = True
            channel = salt.channel.client.AsyncReqChannel.factory(
                self.opts, crypt="clear", io_loop=self.io_loop
            )

        sign_in_payload = self.minion_sign_in_payload()
        try:
            payload = yield channel.send(sign_in_payload, tries=tries, timeout=timeout)
        except SaltReqTimeoutError as e:
            if safe:
                log.warning("SaltReqTimeoutError: %s", e)
                raise salt.ext.tornado.gen.Return("retry")
            if self.opts.get("detect_mode") is True:
                raise salt.ext.tornado.gen.Return("retry")
            else:
                raise SaltClientError(
                    "Attempt to authenticate with the salt master failed with timeout"
                    " error"
                )
        finally:
            if close_channel:
                channel.close()
        ret = self.handle_signin_response(sign_in_payload, payload)
        raise salt.ext.tornado.gen.Return(ret)

    def handle_signin_response(self, sign_in_payload, payload):
        auth = {}
        m_pub_fn = os.path.join(self.opts["pki_dir"], self.mpub)
        auth["master_uri"] = self.opts["master_uri"]
        if not isinstance(payload, dict) or "load" not in payload:
            log.error("Sign-in attempt failed: %s", payload)
            return False
        elif isinstance(payload["load"], dict) and "ret" in payload["load"]:
            if payload["load"]["ret"] == "bad enc algo":
                log.error("Sign-in attempt failed: %s", payload)
                return "bad enc algo"
            elif payload["load"]["ret"] == "bad sig algo":
                log.error("Sign-in attempt failed: %s", payload)
                return "bad sig algo"

        clear_signed_data = payload["load"]
        clear_signature = payload["sig"]
        payload = salt.payload.loads(clear_signed_data)
        if "pub_key" in payload:
            auth["aes"] = self.verify_master(
                payload, master_pub="token" in sign_in_payload
            )
            if not auth["aes"]:
                log.critical(
                    "The Salt Master server's public key did not authenticate!\n"
                    "The master may need to be updated if it is a version of Salt "
                    "lower than %s, or\n"
                    "If you are confident that you are connecting to a valid Salt "
                    "Master, then remove the master public key and restart the "
                    "Salt Minion.\nThe master public key can be found "
                    "at:\n%s",
                    salt.version.__version__,
                    m_pub_fn,
                )
                raise SaltClientError("Invalid master key")

            key = self.get_keys()
            auth["session"] = key.decrypt(
                payload["session"], self.opts["encryption_algorithm"]
            )

        master_pubkey_path = os.path.join(self.opts["pki_dir"], self.mpub)
        if os.path.exists(master_pubkey_path) and not verify_signature(
            master_pubkey_path,
            clear_signed_data,
            clear_signature,
            algorithm=self.opts["signing_algorithm"],
        ):
            log.critical("The payload signature did not validate.")
            raise SaltClientError("Invalid signature")

        if payload["nonce"] != sign_in_payload["nonce"]:
            log.critical("The payload nonce did not validate.")
            raise SaltClientError("Invalid nonce")

        if "ret" in payload:
            if not payload["ret"]:
                if self.opts["rejected_retry"]:
                    log.error(
                        "The Salt Master has rejected this minion's public "
                        "key.\nTo repair this issue, delete the public key "
                        "for this minion on the Salt Master.\nThe Salt "
                        "Minion will attempt to re-authenicate."
                    )
                    return "retry"
                else:
                    log.critical(
                        "The Salt Master has rejected this minion's public "
                        "key!\nTo repair this issue, delete the public key "
                        "for this minion on the Salt Master and restart this "
                        "minion.\nOr restart the Salt Master in open mode to "
                        "clean out the keys. The Salt Minion will now exit."
                    )
                    # Add a random sleep here for systems that are using a
                    # service manager to immediately restart the service to
                    # avoid overloading the system
                    time.sleep(random.randint(10, 20))
                    sys.exit(salt.defaults.exitcodes.EX_NOPERM)
            # Has the master returned that it's maxed out with minions?
            elif payload["ret"] == "full":
                return "full"
            else:
                log.error(
                    "The Salt Master has cached the public key for this "
                    "node, this salt minion will wait for %s seconds "
                    "before attempting to re-authenticate",
                    self.opts["acceptance_wait_time"],
                )
                return "retry"

        if self.opts.get("syndic_master", False):  # Is syndic
            syndic_finger = self.opts.get(
                "syndic_finger", self.opts.get("master_finger", False)
            )
            if syndic_finger:
                if (
                    salt.utils.crypt.pem_finger(
                        m_pub_fn, sum_type=self.opts["hash_type"]
                    )
                    != syndic_finger
                ):
                    self._finger_fail(syndic_finger, m_pub_fn)
        else:
            if self.opts.get("master_finger", False):
                if (
                    salt.utils.crypt.pem_finger(
                        m_pub_fn, sum_type=self.opts["hash_type"]
                    )
                    != self.opts["master_finger"]
                ):
                    self._finger_fail(self.opts["master_finger"], m_pub_fn)

        auth["publish_port"] = payload["publish_port"]
        return auth

    def get_keys(self):
        """
        Return keypair object for the minion.

        :rtype: Crypto.PublicKey.RSA._RSAobj
        :return: The RSA keypair
        """
        if self._private_key is None:
            # Make sure all key parent directories are accessible
            user = self.opts.get("user", "root")
            salt.utils.verify.check_path_traversal(self.opts["pki_dir"], user)

            if not os.path.exists(self.rsa_path):
                log.info("Generating keys: %s", self.opts["pki_dir"])
                gen_keys(
                    self.opts["pki_dir"],
                    "minion",
                    self.opts["keysize"],
                    self.opts.get("user"),
                )
            self._private_key = PrivateKey(self.rsa_path, None)
        return self._private_key

    @salt.utils.decorators.memoize
    def _gen_token(self, key, token):
        return private_encrypt(key, token)

    def gen_token(self, clear_tok):
        """
        Encrypt a string with the minion private key to verify identity
        with the master.

        :param str clear_tok: A plaintext token to encrypt
        :return: Encrypted token
        :rtype: str
        """
        return self._gen_token(self.get_keys(), clear_tok)

    def minion_sign_in_payload(self):
        """
        Generates the payload used to authenticate with the master
        server. This payload consists of the passed in id_ and the ssh
        public key to encrypt the AES key sent back from the master.

        :return: Payload dictionary
        :rtype: dict
        """
        payload = {}
        payload["cmd"] = "_auth"
        payload["id"] = self.opts["id"]
        payload["nonce"] = uuid.uuid4().hex
        payload["enc_algo"] = self.opts["encryption_algorithm"]
        payload["sig_algo"] = self.opts["signing_algorithm"]
        if "autosign_grains" in self.opts:
            autosign_grains = {}
            for grain in self.opts["autosign_grains"]:
                autosign_grains[grain] = self.opts["grains"].get(grain, None)
            payload["autosign_grains"] = autosign_grains
        try:
            pubkey_path = os.path.join(self.opts["pki_dir"], self.mpub)
            pub = PublicKey(pubkey_path)
            payload["token"] = pub.encrypt(
                self.token, self.opts["encryption_algorithm"]
            )
        except FileNotFoundError:
            log.debug("Master public key not found")
        except Exception as exc:  # pylint: disable=broad-except
            log.debug("Exception while encrypting token %s", exc)
        with salt.utils.files.fopen(self.pub_path) as f:
            payload["pub"] = clean_key(f.read())
        return payload

    def decrypt_aes(self, payload, master_pub=True):
        """
        This function is used to decrypt the AES seed phrase returned from
        the master server. The seed phrase is decrypted with the SSH RSA
        host key.

        Pass in the encrypted AES key.
        Returns the decrypted AES seed key, a string

        :param dict payload: The incoming payload. This is a dictionary which may have the following keys:
            'aes': The shared AES key
            'enc': The format of the message. ('clear', 'pub', etc)
            'sig': The message signature
            'publish_port': The TCP port which published the message
            'token': The encrypted token used to verify the message.
            'pub_key': The public key of the sender.

        :rtype: str
        :return: The decrypted token that was provided, with padding.

        :rtype: str
        :return: The decrypted AES seed key
        """
        if self.opts.get("auth_trb", False):
            log.warning("Auth Called: %s", "".join(traceback.format_stack()))
        else:
            log.debug("Decrypting the current master AES key")

        key = self.get_keys()
        key_str = key.decrypt(payload["aes"], self.opts["encryption_algorithm"])
        if "sig" in payload:
            m_path = os.path.join(self.opts["pki_dir"], self.mpub)
            if os.path.exists(m_path):
                try:
                    mkey = PublicKey(m_path)
                except Exception:  # pylint: disable=broad-except
                    return "", ""
                digest = hashlib.sha256(key_str).hexdigest()
                digest = salt.utils.stringutils.to_bytes(digest)
                m_digest = mkey.decrypt(payload["sig"])
                if m_digest != digest:
                    return "", ""
        else:
            return "", ""

        key_str = salt.utils.stringutils.to_str(key_str)

        if "_|-" in key_str:
            return key_str.split("_|-")
        else:
            if "token" in payload:
                token = key.decrypt(payload["token"], self.opts["encryption_algorithm"])
                return key_str, token
            elif not master_pub:
                return key_str, ""
        return "", ""

    def verify_pubkey_sig(self, message, sig):
        """
        Wraps the verify_signature method so we have
        additional checks.

        :rtype: bool
        :return: Success or failure of public key verification
        """
        if self.opts["master_sign_key_name"]:
            path = os.path.join(
                self.opts["pki_dir"], self.opts["master_sign_key_name"] + ".pub"
            )

            if os.path.isfile(path):
                res = verify_signature(
                    path,
                    message,
                    binascii.a2b_base64(sig),
                    algorithm=self.opts["signing_algorithm"],
                )
            else:
                log.error(
                    "Verification public key %s does not exist. You need to "
                    "copy it from the master to the minions pki directory",
                    os.path.basename(path),
                )
                return False
            if res:
                log.debug(
                    "Successfully verified signature of master public key "
                    "with verification public key %s",
                    self.opts["master_sign_key_name"] + ".pub",
                )
                return True
            else:
                log.debug("Failed to verify signature of public key")
                return False
        else:
            log.error(
                "Failed to verify the signature of the message because the "
                "verification key-pairs name is not defined. Please make "
                "sure that master_sign_key_name is defined."
            )
            return False

    def verify_signing_master(self, payload):
        try:
            if self.verify_pubkey_sig(payload["pub_key"], payload["pub_sig"]):
                log.info(
                    "Received signed and verified master pubkey from master %s",
                    self.opts["master"],
                )
                m_pub_fn = os.path.join(self.opts["pki_dir"], self.mpub)
                uid = salt.utils.user.get_uid(self.opts.get("user", None))
                with salt.utils.files.fpopen(m_pub_fn, "wb+", uid=uid) as wfh:
                    wfh.write(salt.utils.stringutils.to_bytes(payload["pub_key"]))
                return True
            else:
                log.error(
                    "Received signed public-key from master %s but signature "
                    "verification failed!",
                    self.opts["master"],
                )
                return False
        except Exception as sign_exc:  # pylint: disable=broad-except
            log.error(
                "There was an error while verifying the masters public-key signature"
            )
            raise Exception(sign_exc)

    def check_auth_deps(self, payload):
        """
        Checks if both master and minion either sign (master) and
        verify (minion). If one side does not, it should fail.

        :param dict payload: The incoming payload. This is a dictionary which may have the following keys:
            'aes': The shared AES key
            'enc': The format of the message. ('clear', 'pub', 'aes')
            'publish_port': The TCP port which published the message
            'token': The encrypted token used to verify the message.
            'pub_key': The RSA public key of the sender.
        """
        # master and minion sign and verify
        if "pub_sig" in payload and self.opts["verify_master_pubkey_sign"]:
            return True
        # master and minion do NOT sign and do NOT verify
        elif "pub_sig" not in payload and not self.opts["verify_master_pubkey_sign"]:
            return True

        # master signs, but minion does NOT verify
        elif "pub_sig" in payload and not self.opts["verify_master_pubkey_sign"]:
            log.error(
                "The masters sent its public-key signature, but signature "
                "verification is not enabled on the minion. Either enable "
                "signature verification on the minion or disable signing "
                "the public key on the master!"
            )
            return False
        # master does NOT sign but minion wants to verify
        elif "pub_sig" not in payload and self.opts["verify_master_pubkey_sign"]:
            log.error(
                "The master did not send its public-key signature, but "
                "signature verification is enabled on the minion. Either "
                "disable signature verification on the minion or enable "
                "signing the public on the master!"
            )
            return False

    def extract_aes(self, payload, master_pub=True):
        """
        Return the AES key received from the master after the minion has been
        successfully authenticated.

        :param dict payload: The incoming payload. This is a dictionary which may have the following keys:
            'aes': The shared AES key
            'enc': The format of the message. ('clear', 'pub', etc)
            'publish_port': The TCP port which published the message
            'token': The encrypted token used to verify the message.
            'pub_key': The RSA public key of the sender.

        :rtype: str
        :return: The shared AES key received from the master.
        """
        if master_pub:
            try:
                aes, token = self.decrypt_aes(payload, master_pub)
                if token != self.token:
                    log.error("The master failed to decrypt the random minion token")
                    return ""
            except Exception:  # pylint: disable=broad-except
                log.error("The master failed to decrypt the random minion token")
                return ""
            return aes
        else:
            aes, token = self.decrypt_aes(payload, master_pub)
            return aes

    def verify_master(self, payload, master_pub=True):
        """
        Verify that the master is the same one that was previously accepted.

        :param dict payload: The incoming payload. This is a dictionary which may have the following keys:
            'aes': The shared AES key
            'enc': The format of the message. ('clear', 'pub', etc)
            'publish_port': The TCP port which published the message
            'token': The encrypted token used to verify the message.
            'pub_key': The RSA public key of the sender.
        :param bool master_pub: Operate as if minion had no master pubkey when it sent auth request, i.e. don't verify
        the minion signature

        :rtype: str
        :return: An empty string on verification failure. On success, the decrypted AES message in the payload.
        """
        m_pub_fn = os.path.join(self.opts["pki_dir"], self.mpub)
        m_pub_exists = os.path.isfile(m_pub_fn)
        if m_pub_exists and master_pub and not self.opts["open_mode"]:
            with salt.utils.files.fopen(m_pub_fn) as fp_:
                local_master_pub = clean_key(fp_.read())

            if payload["pub_key"] != local_master_pub:
                if not self.check_auth_deps(payload):
                    return ""

                if self.opts["verify_master_pubkey_sign"]:
                    if self.verify_signing_master(payload):
                        return self.extract_aes(payload, master_pub=False)
                    else:
                        return ""
                else:
                    # This is not the last master we connected to
                    log.error(
                        "The master key has changed, the salt master could "
                        "have been subverted, verify salt master's public "
                        "key"
                    )
                    return ""

            else:
                if not self.check_auth_deps(payload):
                    return ""
                # verify the signature of the pubkey even if it has
                # not changed compared with the one we already have
                if self.opts["always_verify_signature"]:
                    if self.verify_signing_master(payload):
                        return self.extract_aes(payload)
                    else:
                        log.error(
                            "The masters public could not be verified. Is the "
                            "verification pubkey %s up to date?",
                            self.opts["master_sign_key_name"] + ".pub",
                        )
                        return ""

                else:
                    return self.extract_aes(payload)
        else:
            if not self.check_auth_deps(payload):
                return ""

            # verify the masters pubkey signature if the minion
            # has not received any masters pubkey before
            if self.opts["verify_master_pubkey_sign"]:
                if self.verify_signing_master(payload):
                    return self.extract_aes(payload, master_pub=False)
                else:
                    return ""
            else:
                if not m_pub_exists:
                    # the minion has not received any masters pubkey yet, write
                    # the newly received pubkey to minion_master.pub
                    with salt.utils.files.fopen(m_pub_fn, "wb+") as fp_:
                        fp_.write(salt.utils.stringutils.to_bytes(payload["pub_key"]))
                return self.extract_aes(payload, master_pub=False)

    def _finger_fail(self, finger, master_key):
        log.critical(
            "The specified fingerprint in the master configuration "
            "file:\n%s\nDoes not match the authenticating master's "
            "key:\n%s\nVerify that the configured fingerprint "
            "matches the fingerprint of the correct master and that "
            "this minion is not subject to a man-in-the-middle attack.",
            finger,
            salt.utils.crypt.pem_finger(master_key, sum_type=self.opts["hash_type"]),
        )
        sys.exit(42)


# TODO: remove, we should just return a sync wrapper of AsyncAuth
class SAuth(AsyncAuth):
    """
    Set up an object to maintain authentication with the salt master
    """

    # This class is only a singleton per minion/master pair
    instances = weakref.WeakValueDictionary()

    def __new__(cls, opts, io_loop=None):
        """
        Only create one instance of SAuth per __key()
        """
        key = cls.__key(opts)
        auth = SAuth.instances.get(key)
        if auth is None:
            log.debug("Initializing new SAuth for %s", key)
            auth = object.__new__(cls)
            auth.__singleton_init__(opts)
            SAuth.instances[key] = auth
        else:
            log.debug("Re-using SAuth for %s", key)
        return auth

    @classmethod
    def __key(cls, opts, io_loop=None):
        return (
            opts["pki_dir"],  # where the keys are stored
            opts["id"],  # minion ID
            opts["master_uri"],  # master ID
        )

    # has to remain empty for singletons, since __init__ will *always* be called
    def __init__(self, opts, io_loop=None):
        super().__init__(opts, io_loop=io_loop)

    # an init for the singleton instance to call
    def __singleton_init__(self, opts, io_loop=None):
        """
        Init an Auth instance

        :param dict opts: Options for this server
        :return: Auth instance
        :rtype: Auth
        """
        self.opts = opts
        self.token = salt.utils.stringutils.to_bytes(Crypticle.generate_key_string())
        self.pub_path = os.path.join(self.opts["pki_dir"], "minion.pub")
        self.rsa_path = os.path.join(self.opts["pki_dir"], "minion.pem")
        self._private_key = None
        self._creds = None
        if "syndic_master" in self.opts:
            self.mpub = "syndic_master.pub"
        elif "alert_master" in self.opts:
            self.mpub = "monitor_master.pub"
        else:
            self.mpub = "minion_master.pub"
        if not os.path.isfile(self.pub_path):
            self.get_keys()

    @property
    def creds(self):
        if not hasattr(self, "_creds"):
            self.authenticate()
        return self._creds

    @property
    def crypticle(self):
        if not hasattr(self, "_crypticle"):
            self.authenticate()
        return self._crypticle

    def authenticate(self, _=None):  # TODO: remove unused var
        """
        Authenticate with the master, this method breaks the functional
        paradigm, it will update the master information from a fresh sign
        in, signing in can occur as often as needed to keep up with the
        revolving master AES key.

        :rtype: Crypticle
        :returns: A crypticle used for encryption operations
        """
        acceptance_wait_time = self.opts["acceptance_wait_time"]
        acceptance_wait_time_max = self.opts["acceptance_wait_time_max"]
        if not acceptance_wait_time_max:
            acceptance_wait_time_max = acceptance_wait_time
        with salt.channel.client.ReqChannel.factory(
            self.opts, crypt="clear"
        ) as channel:
            while True:
                creds = self.sign_in(channel=channel)
                if creds == "retry":
                    if self.opts.get("caller"):
                        # We have a list of masters, so we should break
                        # and try the next one in the list.
                        if self.opts.get("local_masters", None):
                            error = SaltClientError(
                                "Minion failed to authenticate"
                                " with the master, has the "
                                "minion key been accepted?"
                            )
                            break
                        else:
                            print(
                                "Minion failed to authenticate with the master, "
                                "has the minion key been accepted?"
                            )
                            sys.exit(2)
                    if acceptance_wait_time:
                        log.info(
                            "Waiting %s seconds before retry.", acceptance_wait_time
                        )
                        time.sleep(acceptance_wait_time)
                    if acceptance_wait_time < acceptance_wait_time_max:
                        acceptance_wait_time += acceptance_wait_time
                        log.debug(
                            "Authentication wait time is %s", acceptance_wait_time
                        )
                    continue
                break
            new_aes, changed_aes, changed_session = False, False, False
            if self._creds is None:
                new_aes = True
                log.error("%s Got new master aes key.", self)
            else:
                if self._creds["aes"] != creds["aes"]:
                    changed_aes = True
                    log.debug("%s The master's aes key has changed.", self)
                if self._creds["session"] != creds["session"]:
                    changed_session = True
                    log.debug("%s The master's session key has changed.", self)
            if new_aes or changed_aes or changed_session:
                self._creds = creds
                self._crypticle = Crypticle(self.opts, creds["aes"])
                self._session_crypticle = Crypticle(self.opts, creds["session"])

    def sign_in(self, timeout=60, safe=True, tries=1, channel=None):
        """
        Send a sign in request to the master, sets the key information and
        returns a dict containing the master publish interface to bind to
        and the decrypted aes key for transport decryption.

        :param int timeout: Number of seconds to wait before timing out the sign-in request
        :param bool safe: If True, do not raise an exception on timeout. Retry instead.
        :param int tries: The number of times to try to authenticate before giving up.

        :raises SaltReqTimeoutError: If the sign-in request has timed out and :param safe: is not set

        :return: Return a string on failure indicating the reason for failure. On success, return a dictionary
        with the publication port and the shared AES key.

        """
        auth = {}

        auth_timeout = self.opts.get("auth_timeout", None)
        if auth_timeout is not None:
            timeout = auth_timeout
        auth_safemode = self.opts.get("auth_safemode", None)
        if auth_safemode is not None:
            safe = auth_safemode
        auth_tries = self.opts.get("auth_tries", None)
        if auth_tries is not None:
            tries = auth_tries

        m_pub_fn = os.path.join(self.opts["pki_dir"], self.mpub)

        auth["master_uri"] = self.opts["master_uri"]

        close_channel = False
        if not channel:
            close_channel = True
            channel = salt.channel.client.ReqChannel.factory(self.opts, crypt="clear")

        sign_in_payload = self.minion_sign_in_payload()
        try:
            payload = channel.send(sign_in_payload, tries=tries, timeout=timeout)
        except SaltReqTimeoutError as e:
            if safe:
                log.warning("SaltReqTimeoutError: %s", e)
                return "retry"
            raise SaltClientError(
                "Attempt to authenticate with the salt master failed with timeout error"
            )
        finally:
            if close_channel:
                channel.close()

        return self.handle_signin_response(sign_in_payload, payload)


class Crypticle:
    """
    Authenticated encryption class

    Encryption algorithm: AES-CBC
    Signing algorithm: HMAC-SHA256
    """

    PICKLE_PAD = b"pickle::"
    AES_BLOCK_SIZE = 16
    SIG_SIZE = hashlib.sha256().digest_size

    def __init__(self, opts, key_string, key_size=192, serial=0):
        self.key_string = key_string
        self.keys = self.extract_keys(self.key_string, key_size)
        self.key_size = key_size
        self.serial = serial

    @classmethod
    def generate_key_string(cls, key_size=192):
        key = os.urandom(key_size // 8 + cls.SIG_SIZE)
        b64key = base64.b64encode(key)
        b64key = b64key.decode("utf-8")
        # Return data must be a base64-encoded string, not a unicode type
        return b64key.replace("\n", "")

    @classmethod
    def write_key(cls, path, key_size=192):
        directory = pathlib.Path(path).parent
        with salt.utils.files.set_umask(0o177):
            fd, tmp = tempfile.mkstemp(dir=directory, prefix="aes")
            os.close(fd)
            with salt.utils.files.fopen(tmp, "w") as fp:
                fp.write(cls.generate_key_string(key_size))
            os.rename(tmp, path)

    @classmethod
    def read_key(cls, path):
        try:
            with salt.utils.files.fopen(path, "r") as fp:
                return fp.read()
        except FileNotFoundError:
            pass

    @classmethod
    def extract_keys(cls, key_string, key_size):
        key = salt.utils.stringutils.to_bytes(base64.b64decode(key_string))
        assert len(key) == key_size / 8 + cls.SIG_SIZE, "invalid key"
        return key[: -cls.SIG_SIZE], key[-cls.SIG_SIZE :]

    def encrypt(self, data):
        """
        encrypt data with AES-CBC and sign it with HMAC-SHA256
        """
        aes_key, hmac_key = self.keys
        pad = self.AES_BLOCK_SIZE - len(data) % self.AES_BLOCK_SIZE
        data = data + salt.utils.stringutils.to_bytes(pad * chr(pad))
        iv_bytes = os.urandom(self.AES_BLOCK_SIZE)
        cipher = Cipher(algorithms.AES(aes_key), modes.CBC(iv_bytes))
        encryptor = cipher.encryptor()
        encr = encryptor.update(data)
        encr += encryptor.finalize()
        data = iv_bytes + encr
        sig = hmac.new(hmac_key, data, hashlib.sha256).digest()
        return data + sig

    def decrypt(self, data):
        """
        verify HMAC-SHA256 signature and decrypt data with AES-CBC
        """
        aes_key, hmac_key = self.keys
        sig = data[-self.SIG_SIZE :]
        data = data[: -self.SIG_SIZE]
        if not isinstance(data, bytes):
            data = salt.utils.stringutils.to_bytes(data)
        mac_bytes = hmac.new(hmac_key, data, hashlib.sha256).digest()
        if len(mac_bytes) != len(sig):
            log.debug("Failed to authenticate message")
            raise AuthenticationError("message authentication failed")
        result = 0
        for zipped_x, zipped_y in zip(mac_bytes, sig):
            result |= zipped_x ^ zipped_y
        if result != 0:
            log.debug("Failed to authenticate message")
            raise AuthenticationError("message authentication failed")
        iv_bytes = data[: self.AES_BLOCK_SIZE]
        data = data[self.AES_BLOCK_SIZE :]
        cipher = Cipher(algorithms.AES(aes_key), modes.CBC(iv_bytes))
        decryptor = cipher.decryptor()
        data = decryptor.update(data) + decryptor.finalize()
        return data[: -data[-1]]

    def dumps(self, obj, nonce=None):
        """
        Serialize and encrypt a python object
        """
        if nonce:
            toencrypt = self.PICKLE_PAD + nonce.encode() + salt.payload.dumps(obj)
        else:
            toencrypt = self.PICKLE_PAD + salt.payload.dumps(obj)
        return self.encrypt(toencrypt)

    def loads(self, data, raw=False, nonce=None):
        """
        Decrypt and un-serialize a python object
        """
        data = self.decrypt(data)
        # simple integrity check to verify that we got meaningful data
        if not data.startswith(self.PICKLE_PAD):
            return {}
        data = data[len(self.PICKLE_PAD) :]
        if nonce:
            ret_nonce = data[:32].decode()
            data = data[32:]
            if ret_nonce != nonce:
                raise SaltClientError("Nonce verification error")
        payload = salt.payload.loads(data, raw=raw)
        if isinstance(payload, dict):
            if "serial" in payload:
                serial = payload.pop("serial")
                if serial <= self.serial:
                    log.critical(
                        "A message with an invalid serial was received.\n"
                        "this serial: %d\n"
                        "last serial: %d\n"
                        "The minion will not honor this request.",
                        serial,
                        self.serial,
                    )
                    return {}
                self.serial = serial
        return payload



"""
This is the default grains PCRE matcher.
"""

import logging

import salt.utils.data
from salt.defaults import DEFAULT_TARGET_DELIM

log = logging.getLogger(__name__)


def match(tgt, delimiter=DEFAULT_TARGET_DELIM, opts=None, minion_id=None):
    """
    Matches a grain based on regex
    """
    if not opts:
        opts = __opts__
    log.debug("grains pcre target: %s", tgt)
    if delimiter not in tgt:
        log.error(
            "Got insufficient arguments for grains pcre match statement from master"
        )
        return False

    return salt.utils.data.subdict_match(
        opts["grains"], tgt, delimiter=delimiter, regex_match=True
    )



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""

import os

import pytest

import salt.modules.qemu_img as qemu_img
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {qemu_img: {}}


def test_make_image():
    """
    Test for create a blank virtual machine image file
    of the specified size in megabytes
    """
    with patch.object(
        os.path, "isabs", MagicMock(side_effect=[False, True, True, True])
    ):
        assert qemu_img.make_image("location", "size", "fmt") == ""

        with patch.object(os.path, "isdir", MagicMock(side_effect=[False, True, True])):
            assert qemu_img.make_image("location", "size", "fmt") == ""

            with patch.dict(
                qemu_img.__salt__,
                {"cmd.retcode": MagicMock(side_effect=[False, True])},
            ):
                assert qemu_img.make_image("location", "size", "fmt") == "location"
                assert qemu_img.make_image("location", "size", "fmt") == ""



"""
Spacewalk Runner
================

.. versionadded:: 2016.3.0

Runner to interact with Spacewalk using Spacewalk API

:codeauthor: Nitin Madhok <nmadhok@g.clemson.edu>, Joachim Werner <joe@suse.com>, Benedikt Werner <1benediktwerner@gmail.com>
:maintainer: Benedikt Werner <1benediktwerner@gmail.com>

To use this runner, set up the Spacewalk URL, username and password in the
master configuration at ``/etc/salt/master`` or ``/etc/salt/master.d/spacewalk.conf``:

.. code-block:: yaml

    spacewalk:
      spacewalk01.domain.com:
        username: 'testuser'
        password: 'verybadpass'
      spacewalk02.domain.com:
        username: 'testuser'
        password: 'verybadpass'

.. note::

    Optionally, ``protocol`` can be specified if the spacewalk server is
    not using the defaults. Default is ``protocol: https``.

"""

import atexit
import logging
import xmlrpc.client  # nosec

log = logging.getLogger(__name__)

_sessions = {}


def __virtual__():
    """
    Check for spacewalk configuration in master config file
    or directory and load runner only if it is specified
    """
    if not _get_spacewalk_configuration():
        return False, "No spacewalk configuration found"
    return True


def _get_spacewalk_configuration(spacewalk_url=""):
    """
    Return the configuration read from the master configuration
    file or directory
    """
    spacewalk_config = __opts__["spacewalk"] if "spacewalk" in __opts__ else None

    if spacewalk_config:
        try:
            for spacewalk_server, service_config in spacewalk_config.items():
                username = service_config.get("username", None)
                password = service_config.get("password", None)
                protocol = service_config.get("protocol", "https")

                if not username or not password:
                    log.error(
                        "Username or Password has not been specified in the master "
                        "configuration for %s",
                        spacewalk_server,
                    )
                    return False

                ret = {
                    "api_url": f"{protocol}://{spacewalk_server}/rpc/api",
                    "username": username,
                    "password": password,
                }

                if (not spacewalk_url) or (spacewalk_url == spacewalk_server):
                    return ret
        except Exception as exc:  # pylint: disable=broad-except
            log.error("Exception encountered: %s", exc)
            return False

        if spacewalk_url:
            log.error(
                "Configuration for %s has not been specified in the master "
                "configuration",
                spacewalk_url,
            )
            return False

    return False


def _get_client_and_key(url, user, password, verbose=0):
    """
    Return the client object and session key for the client
    """
    session = {}
    session["client"] = xmlrpc.client.Server(url, verbose=verbose, use_datetime=True)
    session["key"] = session["client"].auth.login(user, password)

    return session


def _disconnect_session(session):
    """
    Disconnect API connection
    """
    session["client"].auth.logout(session["key"])


def _get_session(server):
    """
    Get session and key
    """
    if server in _sessions:
        return _sessions[server]

    config = _get_spacewalk_configuration(server)
    if not config:
        raise Exception(f"No config for '{server}' found on master")

    session = _get_client_and_key(
        config["api_url"], config["username"], config["password"]
    )
    atexit.register(_disconnect_session, session)

    client = session["client"]
    key = session["key"]
    _sessions[server] = (client, key)

    return client, key


def api(server, command, *args, **kwargs):
    """
    Call the Spacewalk xmlrpc api.

    CLI Example:

    .. code-block:: bash

        salt-run spacewalk.api spacewalk01.domain.com systemgroup.create MyGroup Description
        salt-run spacewalk.api spacewalk01.domain.com systemgroup.create arguments='["MyGroup", "Description"]'

    State Example:

    .. code-block:: yaml

        create_group:
          salt.runner:
            - name: spacewalk.api
            - server: spacewalk01.domain.com
            - command: systemgroup.create
            - arguments:
              - MyGroup
              - Description
    """
    if "arguments" in kwargs:
        arguments = kwargs["arguments"]
    else:
        arguments = args

    call = f"{command} {arguments}"
    try:
        client, key = _get_session(server)
    except Exception as exc:  # pylint: disable=broad-except
        err_msg = (
            "Exception raised when connecting to spacewalk server ({}): {}".format(
                server, exc
            )
        )
        log.error(err_msg)
        return {call: err_msg}

    namespace, _, method = command.rpartition(".")
    if not namespace:
        return {
            call: "Error: command must use the following format: 'namespace.method'"
        }
    endpoint = getattr(getattr(client, namespace), method)

    try:
        output = endpoint(key, *arguments)
    except Exception as e:  # pylint: disable=broad-except
        output = f"API call failed: {e}"

    return {call: output}


def addGroupsToKey(server, activation_key, groups):
    """
    Add server groups to a activation key

    CLI Example:

    .. code-block:: bash

        salt-run spacewalk.addGroupsToKey spacewalk01.domain.com 1-my-key '[group1, group2]'
    """

    try:
        client, key = _get_session(server)
    except Exception as exc:  # pylint: disable=broad-except
        err_msg = (
            "Exception raised when connecting to spacewalk server ({}): {}".format(
                server, exc
            )
        )
        log.error(err_msg)
        return {"Error": err_msg}

    all_groups = client.systemgroup.listAllGroups(key)
    groupIds = []
    for group in all_groups:
        if group["name"] in groups:
            groupIds.append(group["id"])

    if client.activationkey.addServerGroups(key, activation_key, groupIds) == 1:
        return {activation_key: groups}
    else:
        return {activation_key: "Failed to add groups to activation key"}


def deleteAllGroups(server):
    """
    Delete all server groups from Spacewalk
    """

    try:
        client, key = _get_session(server)
    except Exception as exc:  # pylint: disable=broad-except
        err_msg = (
            "Exception raised when connecting to spacewalk server ({}): {}".format(
                server, exc
            )
        )
        log.error(err_msg)
        return {"Error": err_msg}

    groups = client.systemgroup.listAllGroups(key)

    deleted_groups = []
    failed_groups = []

    for group in groups:
        if client.systemgroup.delete(key, group["name"]) == 1:
            deleted_groups.append(group["name"])
        else:
            failed_groups.append(group["name"])

    ret = {"deleted": deleted_groups}
    if failed_groups:
        ret["failed"] = failed_groups
    return ret


def deleteAllSystems(server):
    """
    Delete all systems from Spacewalk

    CLI Example:

    .. code-block:: bash

        salt-run spacewalk.deleteAllSystems spacewalk01.domain.com
    """

    try:
        client, key = _get_session(server)
    except Exception as exc:  # pylint: disable=broad-except
        err_msg = (
            "Exception raised when connecting to spacewalk server ({}): {}".format(
                server, exc
            )
        )
        log.error(err_msg)
        return {"Error": err_msg}

    systems = client.system.listSystems(key)

    ids = []
    names = []
    for system in systems:
        ids.append(system["id"])
        names.append(system["name"])

    if client.system.deleteSystems(key, ids) == 1:
        return {"deleted": names}
    else:
        return {"Error": "Failed to delete all systems"}


def deleteAllActivationKeys(server):
    """
    Delete all activation keys from Spacewalk

    CLI Example:

    .. code-block:: bash

        salt-run spacewalk.deleteAllActivationKeys spacewalk01.domain.com
    """

    try:
        client, key = _get_session(server)
    except Exception as exc:  # pylint: disable=broad-except
        err_msg = (
            "Exception raised when connecting to spacewalk server ({}): {}".format(
                server, exc
            )
        )
        log.error(err_msg)
        return {"Error": err_msg}

    activation_keys = client.activationkey.listActivationKeys(key)

    deleted_keys = []
    failed_keys = []

    for aKey in activation_keys:
        if client.activationkey.delete(key, aKey["key"]) == 1:
            deleted_keys.append(aKey["key"])
        else:
            failed_keys.append(aKey["key"])

    ret = {"deleted": deleted_keys}
    if failed_keys:
        ret["failed"] = failed_keys
    return ret


def unregister(name, server_url):
    """
    Unregister specified server from Spacewalk

    CLI Example:

    .. code-block:: bash

        salt-run spacewalk.unregister my-test-vm spacewalk01.domain.com
    """

    try:
        client, key = _get_session(server_url)
    except Exception as exc:  # pylint: disable=broad-except
        err_msg = (
            "Exception raised when connecting to spacewalk server ({}): {}".format(
                server_url, exc
            )
        )
        log.error(err_msg)
        return {name: err_msg}

    systems_list = client.system.getId(key, name)

    if systems_list:
        for system in systems_list:
            out = client.system.deleteSystem(key, system["id"])
            if out == 1:
                return {name: f"Successfully unregistered from {server_url}"}
            else:
                return {name: f"Failed to unregister from {server_url}"}
    else:
        return {name: f"System does not exist in spacewalk server ({server_url})"}



"""
tests for user state
user absent
user present
user present with custom homedir
"""

import pathlib
import random
import shutil
import sys

import pytest
from saltfactories.utils import random_string

import salt.utils.files
import salt.utils.platform

try:
    import grp
except ImportError:
    grp = None

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_not_root,
    pytest.mark.destructive_test,
    pytest.mark.windows_whitelisted,
]

ANSI_FILESYSTEM_ENCODING = sys.getfilesystemencoding().startswith("ANSI")


@pytest.fixture
def username(sminion):
    _username = random_string("new-account-", uppercase=False)
    try:
        yield _username
    finally:
        try:
            sminion.functions.user.delete(_username, remove=True, force=True)
        except Exception:  # pylint: disable=broad-except
            # The point here is just system cleanup. It can fail if no account was created
            pass


@pytest.fixture
def guid():
    return random.randint(60000, 61000)


@pytest.fixture
def user_home(username, tmp_path):
    if salt.utils.platform.is_windows():
        return tmp_path / username

    return pathlib.Path("/var/lib") / username


@pytest.fixture
def group_1(username, grains):
    groupname = username
    if salt.utils.platform.is_darwin():
        groupname = "staff"
    elif salt.utils.platform.is_photonos():
        groupname = "users"
    elif grains["os_family"] in ("Suse",):
        groupname = "users"
    with pytest.helpers.create_group(name=groupname) as group:
        yield group


@pytest.fixture
def group_2():
    with pytest.helpers.create_group() as group:
        yield group


@pytest.fixture
def existing_account():
    with pytest.helpers.create_account(create_group=True) as _account:
        yield _account


@pytest.mark.slow_test
def test_user_absent(states):
    """
    Test user.absent with a non existing account
    """
    ret = states.user.absent(name=random_string("account-", uppercase=False))
    assert ret.result is True


def test_user_absent_existing_account(states, existing_account):
    """
    Test user.absent with an existing account
    """
    ret = states.user.absent(name=existing_account.username)
    assert ret.result is True


def test_user_present(states, username):
    """
    Test user.present with a non existing account
    """
    ret = states.user.present(name=username)
    assert ret.result is True


def test_user_present_with_existing_group(states, username, existing_account):
    ret = states.user.present(username, gid=existing_account.group.info.gid)
    assert ret.result is True


@pytest.mark.skip_on_windows(
    reason="Home directories are handled differently in Windows"
)
def test_user_present_when_home_dir_does_not_18843(states, existing_account):
    """
    User exists but home directory does not. Home directory get's created
    """
    shutil.rmtree(existing_account.info.home)
    ret = states.user.present(
        name=existing_account.username,
        home=existing_account.info.home,
    )
    assert ret.result is True
    assert pathlib.Path(existing_account.info.home).is_dir()


def test_user_present_nondefault(grains, modules, states, username, user_home):
    ret = states.user.present(name=username, home=str(user_home))
    assert ret.result is True

    user_info = modules.user.info(username)
    assert user_info

    if salt.utils.platform.is_windows():
        group_name = modules.user.list_groups(username)
    else:
        group_name = grp.getgrgid(user_info["gid"]).gr_name

    if not salt.utils.platform.is_darwin() and not salt.utils.platform.is_windows():
        assert user_home.is_dir()

    if grains["os_family"] in ("Suse",):
        expected_group_name = "users"
    elif grains["os_family"] == "MacOS":
        expected_group_name = "staff"
    elif salt.utils.platform.is_windows():
        expected_group_name = []
    elif grains["os"] == "VMware Photon OS":
        expected_group_name = "users"
    else:
        expected_group_name = username
    assert group_name == expected_group_name


@pytest.mark.skip_on_windows(reason="windows minion does not support 'usergroup'")
def test_user_present_usergroup_false(modules, states, username, group_1, user_home):
    ret = states.user.present(
        name=username,
        gid=group_1.info.gid,
        usergroup=False,
        home=str(user_home),
    )
    assert ret.result is True

    if not salt.utils.platform.is_darwin():
        assert user_home.is_dir()

    user_info = modules.user.info(username)
    assert user_info

    group_name = grp.getgrgid(user_info["gid"]).gr_name
    assert group_name == group_1.name


@pytest.mark.skip_on_windows(reason="windows minion does not support 'usergroup'")
def test_user_present_usergroup_true(modules, states, username, user_home, group_1):
    ret = states.user.present(
        name=username,
        gid=group_1.info.gid,
        usergroup=True,
        home=str(user_home),
    )
    assert ret.result is True

    if not salt.utils.platform.is_darwin():
        assert user_home.is_dir()

    user_info = modules.user.info(username)
    assert user_info

    group_name = grp.getgrgid(user_info["gid"]).gr_name
    assert group_name == group_1.name


def _check_skip(grains):
    if grains["os"] == "MacOS":
        return True
    return False


@pytest.mark.skipif(
    ANSI_FILESYSTEM_ENCODING,
    reason=(
        "A system encoding which supports Unicode characters must be set. "
        f"Current setting is: {sys.getfilesystemencoding()}. "
        "Try setting $LANG='en_US.UTF-8'"
    ),
)
@pytest.mark.skip_initial_gh_actions_failure(skip=_check_skip)
def test_user_present_unicode(states, username, subtests):
    """
    It ensures that unicode GECOS data will be properly handled, without
    any encoding-related failures.
    """
    with subtests.test("Non existing account"):
        ret = states.user.present(
            name=username,
            fullname="Sålt Test",
            roomnumber="①②③",
            workphone="١٢٣٤",
            homephone="६७८",
        )
        assert ret.result is True

    with subtests.test("Update existing account"):
        ret = states.user.present(
            name=username,
            fullname="Sålt Test",
            roomnumber="①②③",
            workphone="١٢٣٤",
            homephone="६७८",
        )
        assert ret.result is True


@pytest.mark.skip_on_windows(
    reason="windows minion does not support roomnumber or phone",
)
def test_user_present_gecos(modules, states, username):
    """
    It ensures that numeric GECOS data will be properly coerced to strings,
    otherwise the state will fail because the GECOS fields are written as
    strings (and show up in the user.info output as such). Thus the
    comparison will fail, since '12345' != 12345.
    """
    fullname = 123345
    roomnumber = 123
    workphone = homephone = 1234567890
    ret = states.user.present(
        name=username,
        fullname=fullname,
        roomnumber=roomnumber,
        workphone=workphone,
        homephone=homephone,
    )
    assert ret.result is True

    user_info = modules.user.info(username)
    assert user_info

    assert user_info["fullname"] == str(fullname)
    if not salt.utils.platform.is_darwin():
        # MacOS does not supply the following GECOS fields
        assert user_info["roomnumber"] == str(roomnumber)
        assert user_info["workphone"] == str(workphone)
        assert user_info["homephone"] == str(homephone)


@pytest.mark.skip_on_windows(
    reason="windows minion does not support roomnumber or phone",
)
def test_user_present_gecos_empty_fields(modules, states, username):
    """
    It ensures that if no GECOS data is supplied, the fields will be coerced
    into empty strings as opposed to the string "None".
    """
    fullname = roomnumber = workphone = homephone = ""
    ret = states.user.present(
        name=username,
        fullname=fullname,
        roomnumber=roomnumber,
        workphone=workphone,
        homephone=homephone,
    )
    assert ret.result is True

    user_info = modules.user.info(username)
    assert user_info

    assert user_info["fullname"] == fullname
    if not salt.utils.platform.is_darwin():
        # MacOS does not supply the following GECOS fields
        assert user_info["roomnumber"] == roomnumber
        assert user_info["workphone"] == workphone
        assert user_info["homephone"] == homephone


@pytest.mark.skip_on_windows(reason="windows minion does not support createhome")
@pytest.mark.parametrize("createhome", [True, False])
def test_user_present_home_directory_created(modules, states, username, createhome):
    """
    It ensures that the home directory is created.
    """
    ret = states.user.present(name=username, createhome=createhome)
    assert ret.result is True

    user_info = modules.user.info(username)
    assert user_info

    assert pathlib.Path(user_info["home"]).is_dir() is createhome


@pytest.mark.skip_on_windows(reason="windows minion does not support persist_home")
@pytest.mark.skip_on_darwin(reason="Mac minion does not support persist_home")
@pytest.mark.parametrize("persist_home", [True, False])
def test_user_present_home_directory_persisted(
    modules, states, existing_account, user_home, persist_home
):
    """
    It ensures that the home directory is persisted when home changed.
    """
    pathlib.Path(existing_account.info.home, "testfile").touch()
    ret = states.user.present(
        name=existing_account.username, home=str(user_home), persist_home=persist_home
    )
    assert ret.result is True

    user_info = modules.user.info(existing_account.info.name)
    assert user_info

    assert pathlib.Path(user_info["home"], "testfile").is_file() is persist_home


@pytest.mark.skip_on_darwin(reason="groups/gid not fully supported")
@pytest.mark.skip_on_windows(reason="groups/gid not fully supported")
def test_user_present_change_gid_but_keep_group(
    modules, states, username, group_1, group_2
):
    """
    This tests the case in which the default group is changed at the same
    time as it is also moved into the "groups" list.
    """

    # Add the user
    ret = states.user.present(name=username, gid=group_1.info.gid)
    assert ret.result is True

    user_info = modules.user.info(username)
    assert user_info

    assert user_info["gid"] == group_1.info.gid
    assert user_info["groups"] == [group_1.name]

    # Now change the gid and move alt_group to the groups list in the
    # same salt run.
    ret = states.user.present(
        name=username,
        gid=group_2.info.gid,
        groups=[group_1.name],
        allow_gid_change=True,
    )
    assert ret.result is True

    # Be sure that we did what we intended
    user_info = modules.user.info(username)
    assert user_info

    assert user_info["gid"] == group_2.info.gid
    assert user_info["groups"] == [group_2.name, group_1.name]


@pytest.mark.skip_unless_on_windows
def test_user_present_existing(states, username):
    win_profile = f"C:\\User\\{username}"
    win_logonscript = "C:\\logon.vbs"
    win_description = "Test User Account"
    ret = states.user.present(
        name=username,
        win_homedrive="U:",
        win_profile=win_profile,
        win_logonscript=win_logonscript,
        win_description=win_description,
    )
    assert ret.result is True

    win_profile = f"C:\\Users\\{username}"
    win_description = "Temporary Account"
    ret = states.user.present(
        name=username,
        win_homedrive="R:",
        win_profile=win_profile,
        win_logonscript=win_logonscript,
        win_description=win_description,
    )
    assert ret.result is True
    assert ret.changes
    assert "homedrive" in ret.changes
    assert ret.changes["homedrive"] == "R:"
    assert "profile" in ret.changes
    assert ret.changes["profile"] == win_profile
    assert "description" in ret.changes
    assert ret.changes["description"] == win_description


@pytest.mark.skip_unless_on_linux(reason="underlying functionality only runs on Linux")
def test_user_present_change_groups(modules, states, username, group_1, group_2):
    ret = states.user.present(
        name=username,
        groups=[group_1.name, group_2.name],
    )
    assert ret.result is True

    user_info = modules.user.info(username)
    assert user_info
    assert user_info["groups"] == [group_2.name, group_1.name]

    # run again and remove group_2
    ret = states.user.present(
        name=username,
        groups=[group_1.name],
    )
    assert ret.result is True

    user_info = modules.user.info(username)
    assert user_info
    assert user_info["groups"] == [group_1.name]


@pytest.mark.skip_unless_on_linux(reason="underlying functionality only runs on Linux")
def test_user_present_change_optional_groups(
    modules, states, username, group_1, group_2
):
    ret = states.user.present(
        name=username,
        optional_groups=[group_1.name, group_2.name],
    )
    assert ret.result is True

    user_info = modules.user.info(username)
    assert user_info
    assert user_info["groups"] == [group_2.name, group_1.name]

    # run again and remove group_2
    ret = states.user.present(
        name=username,
        optional_groups=[group_1.name],
    )
    assert ret.result is True

    user_info = modules.user.info(username)
    assert user_info
    assert user_info["groups"] == [group_1.name]


@pytest.fixture
def user_present_groups(states):
    groups = ["testgroup1", "testgroup2"]
    try:
        yield groups
    finally:
        for group in groups:
            ret = states.group.absent(name=group)
            assert ret.result is True


@pytest.mark.skip_unless_on_linux(reason="underlying functionality only runs on Linux")
def test_user_present_no_groups(modules, states, username, user_present_groups, guid):
    """
    test user.present when groups arg is not
    included by the group is created in another
    state. Re-run the states to ensure there are
    not changes and it is idempotent.
    """
    ret = states.group.present(name=username, gid=guid)
    assert ret.result is True

    ret = states.user.present(
        name=username,
        uid=guid,
        gid=guid,
    )
    assert ret.result is True
    assert ret.changes["groups"] == [username]
    assert ret.changes["name"] == username

    ret = states.group.present(
        name=user_present_groups[0],
        members=[username],
    )
    assert ret.changes["members"] == [username]

    ret = states.group.present(
        name=user_present_groups[1],
        members=[username],
    )
    assert ret.changes["members"] == [username]

    user_info = modules.user.info(username)
    assert user_info
    assert user_info["groups"] == [username, *user_present_groups]

    # run again, expecting no changes
    ret = states.group.present(name=username)
    assert ret.result is True
    assert ret.changes == {}

    ret = states.user.present(
        name=username,
    )
    assert ret.result is True
    assert ret.changes == {}

    ret = states.group.present(
        name=user_present_groups[0],
        members=[username],
    )
    assert ret.result is True
    assert ret.changes == {}

    ret = states.group.present(
        name=user_present_groups[1],
        members=[username],
    )
    assert ret.result is True
    assert ret.changes == {}

    user_info = modules.user.info(username)
    assert user_info
    assert user_info["groups"] == [username, *user_present_groups]



import glob
import logging
import os
import pathlib
import pprint
import shutil
import sys

import pytest

import salt.utils.files
import salt.utils.path
import salt.utils.platform
import salt.utils.versions
import salt.utils.win_dacl
import salt.utils.win_functions
import salt.utils.win_runas
from tests.support.helpers import (
    SKIP_INITIAL_PHOTONOS_FAILURES,
    VirtualEnv,
    patched_environ,
)

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.skip_on_fips_enabled_platform,
]


def _win_user_where(username, password, program):
    cmd = f"cmd.exe /c where {program}"
    ret = salt.utils.win_runas.runas(cmd, username, password)
    assert ret["retcode"] == 0
    return ret["stdout"].strip().split("\n")[-1].strip()


@pytest.fixture
def pkg_name():
    return "pep8"


@pytest.fixture
def venv(tmp_path):
    with VirtualEnv(venv_dir=tmp_path) as venv:
        yield venv


@pytest.fixture
def account():
    with pytest.helpers.create_account(username="pip-account") as _account:
        yield _account


@pytest.fixture(scope="module")
def create_virtualenv(modules):
    def run_command(path, **kwargs):
        """
        The reason why the virtualenv creation is proxied by this function is mostly
        because under windows, we can't seem to properly create a virtualenv off of
        another virtualenv(we can on linux) and also because, we really don't want to
        test virtualenv creation off of another virtualenv, we want a virtualenv created
        from the original python.
        Also, one windows, we must also point to the virtualenv binary outside the existing
        virtualenv because it will fail otherwise
        """
        if "python" not in kwargs:
            try:
                if salt.utils.platform.is_windows():
                    python = os.path.join(
                        sys.real_prefix, os.path.basename(sys.executable)
                    )
                else:
                    python_binary_names = [
                        "python{}.{}".format(*sys.version_info),
                        "python{}".format(*sys.version_info),
                        "python",
                    ]
                    for binary_name in python_binary_names:
                        python = os.path.join(sys.real_prefix, "bin", binary_name)
                        if os.path.exists(python):
                            break
                    else:
                        pytest.fail(
                            "Couldn't find a python binary name under '{}' matching: {}".format(
                                os.path.join(sys.real_prefix, "bin"),
                                python_binary_names,
                            )
                        )
                # We're running off a virtualenv, and we don't want to create a virtualenv off of
                # a virtualenv, let's point to the actual python that created the virtualenv
                kwargs["python"] = python
            except AttributeError:
                # We're running off of the system python
                pass
        return modules.virtualenv.create(path, **kwargs)

    return run_command


@pytest.fixture
def _skip_if_pep8_installed(modules, pkg_name):
    if pkg_name in modules.pip.list():
        pytest.skip(f"{pkg_name} is already installed, uninstall to run this test")


@pytest.mark.slow_test
@pytest.mark.usefixtures("_skip_if_pep8_installed")
def test_pip_installed_removed(states, pkg_name):
    """
    Tests installed and removed states
    """
    ret = states.pip.installed(name=pkg_name)
    assert ret.result is True
    ret = states.pip.removed(name=pkg_name)
    assert ret.result is True


@pytest.mark.slow_test
def test_pip_installed_removed_venv(states, venv, pkg_name):
    ret = states.pip.installed(name=pkg_name, bin_env=str(venv.venv_dir))
    assert ret.result is True
    ret = states.pip.removed(name=pkg_name, bin_env=str(venv.venv_dir))
    assert ret.result is True


@pytest.mark.slow_test
def test_pip_installed_errors(tmp_path, modules, state_tree):
    venv_dir = tmp_path / "pip-installed-errors"
    # Since we don't have the virtualenv created, pip.installed will
    # throw an error.
    # Example error strings:
    #  * "Error installing 'pep8': /tmp/pip-installed-errors: not found"
    #  * "Error installing 'pep8': /bin/sh: 1: /tmp/pip-installed-errors: not found"
    #  * "Error installing 'pep8': /bin/bash: /tmp/pip-installed-errors: No such file or directory"
    sls_contents = f"""
    pep8-pip:
      pip.installed:
        - name: pep8
        - bin_env: '{venv_dir}'
    """
    with patched_environ(SHELL="/bin/sh"):
        with pytest.helpers.temp_file(
            "pip-installed-errors.sls", sls_contents, state_tree
        ):
            ret = modules.state.sls("pip-installed-errors")
            for state_return in ret:
                assert state_return.result is False
                assert "Error installing 'pep8':" in state_return.comment

            # We now create the missing virtualenv
            with VirtualEnv(venv_dir=venv_dir):
                # The state should not have any issues running now
                ret = modules.state.sls(mods="pip-installed-errors")
                for state_return in ret:
                    assert state_return.result is True


def test_pip_installed_name_test_mode(states, venv, pkg_name):
    """
    Test pip.installed state while test=true
    """
    ret = states.pip.installed(name=pkg_name, bin_env=str(venv.venv_dir), test=True)
    assert pkg_name in ret.comment


def test_pip_installed_pkgs_test_mode(states, venv):
    """
    Test pip.installed state while test=true
    """
    pkgs = ["boto", "pep8", "black"]
    msg = "Python package(s) set to be installed:\nboto\npep8\nblack"
    ret = states.pip.installed(
        name=None, pkgs=pkgs, bin_env=str(venv.venv_dir), test=True
    )
    assert msg in ret.comment


@pytest.mark.slow_test
def test_issue_2028_pip_installed_state(
    tmp_path, modules, state_tree, get_python_executable
):

    venv_dir = tmp_path / "issue-2028-pip-installed"

    sls_contents = """
    {%- set virtualenv_base = salt['pillar.get']('venv_dir') %}
    {%- set python_executable = salt['pillar.get']('python_executable') %}

    {{ virtualenv_base }}:
      virtualenv.managed:
        - system_site_packages: False
        - distribute: False
        {#- Provide the real path for the python executable in case tests are running inside a virtualenv #}
        {%- if python_executable %}
        - python: {{ python_executable }}
        {%- endif %}

    install-working-setuptools:
      pip.installed:
        - name: 'setuptools!=50.*,!=51.*,!=52.*'
        - bin_env: {{ virtualenv_base }}
        - require:
          - virtualenv: {{ virtualenv_base }}

    pep8-pip:
      pip.installed:
        - name: pep8
        - bin_env: {{ virtualenv_base }}
        - require:
          - pip: install-working-setuptools
          - virtualenv: {{ virtualenv_base }}
    """
    with pytest.helpers.temp_file(
        "issue-2028-pip-installed.sls", sls_contents, state_tree
    ):
        ret = modules.state.sls(
            mods="issue-2028-pip-installed",
            pillar={
                "venv_dir": str(venv_dir),
                "python_exeutable": get_python_executable,
            },
        )

        pep8_bin = venv_dir / "bin" / "pep8"

        if salt.utils.platform.is_windows():
            pep8_bin = venv_dir / "Scripts" / "pep8.exe"

        for state_return in ret:
            assert state_return.result is True

        assert os.path.isfile(str(pep8_bin)) is True


@pytest.mark.slow_test
def test_issue_2087_missing_pip(modules, venv, pkg_name, state_tree):
    sls_name = "issue-2087-missing-pip"
    sls_contents = f"""
    {pkg_name}-pip:
      pip.installed:
        - name: {pkg_name}
        - bin_env: {venv.venv_dir}
    """

    with pytest.helpers.temp_file(f"{sls_name}.sls", sls_contents, state_tree):
        # Let's remove the pip binary
        pip_bin = venv.venv_bin_dir / "pip"
        site_dir = pathlib.Path(
            modules.virtualenv.get_distribution_path(str(venv.venv_dir), "pip")
        )
        if salt.utils.platform.is_windows():
            pip_bin = venv.venv_dir / "Scripts" / "pip.exe"
            site_dir = venv.venv_dir / "lib" / "site-packages"
        if not pip_bin.is_file():
            pytest.skip("Failed to find the pip binary to the test virtualenv")
        pip_bin.unlink()

        # Also remove the pip dir from site-packages
        # This is needed now that we're using python -m pip instead of the
        # pip binary directly. python -m pip will still work even if the
        # pip binary is missing
        shutil.rmtree(site_dir / "pip")

        # Let's run the state which should fail because pip is missing
        ret = modules.state.sls(mods=sls_name)
        for state_return in ret:
            assert state_return.result is False
            assert (
                f"Error installing '{pkg_name}': Could not find a `pip` binary"
                in state_return.comment
            )


@SKIP_INITIAL_PHOTONOS_FAILURES
@pytest.mark.destructive_test
@pytest.mark.slow_test
@pytest.mark.skip_if_not_root
def test_issue_6912_wrong_owner(tmp_path, create_virtualenv, states, account):
    # Setup virtual environment directory to be used throughout the test
    venv_dir = tmp_path / "6912-wrong-owner"
    venv_kwargs = {}

    # The virtual environment needs to be in a location that is accessible
    # by both the user running the test and the runas user
    if salt.utils.platform.is_windows():
        salt.utils.win_dacl.set_permissions(tmp_path, account.username, "full_control")
        # Make sure we're calling a virtualenv and python
        # program that the user has access too.
        venv_kwargs["venv_bin"] = _win_user_where(
            account.username,
            account.password,
            "virtualenv",
        )
        venv_kwargs["python"] = _win_user_where(
            account.username,
            account.password,
            "python",
        )
    else:
        os.chown(str(tmp_path), account.info.uid, -1)

    # Create the virtual environment
    venv_create = create_virtualenv(
        str(venv_dir), user=account.username, password=account.password, **venv_kwargs
    )
    if venv_create.get("retcode", 1) > 0:
        pytest.skip(f"Failed to create testcase virtual environment: {venv_create}")

    # pip install passing the package name in `name`
    ret = states.pip.installed(
        name="pep8",
        user=account.username,
        password=account.password,
        bin_env=str(venv_dir),
    )
    assert ret.result is True

    for globmatch in (
        os.path.join(str(venv_dir), "**", "pep8*"),
        os.path.join(str(venv_dir), "*", "**", "pep8*"),
        os.path.join(str(venv_dir), "*", "*", "**", "pep8*"),
    ):
        for path in glob.glob(globmatch):
            if salt.utils.platform.is_windows():
                assert salt.utils.win_dacl.get_owner(path) == account.username
            else:
                assert os.stat(path).st_uid == account.info.uid


@SKIP_INITIAL_PHOTONOS_FAILURES
@pytest.mark.destructive_test
@pytest.mark.skip_on_darwin(reason="Test is flaky on macosx")
@pytest.mark.slow_test
@pytest.mark.skip_if_not_root
def test_issue_6912_wrong_owner_requirements_file(
    tmp_path, create_virtualenv, state_tree, states, account
):
    # Setup virtual environment directory to be used throughout the test
    venv_dir = tmp_path / "6912-wrong-owner"
    venv_kwargs = {}

    # The virtual environment needs to be in a location that is accessible
    # by both the user running the test and the runas user
    if salt.utils.platform.is_windows():
        salt.utils.win_dacl.set_permissions(
            str(tmp_path), account.username, "full_control"
        )
        # Make sure we're calling a virtualenv and python
        # program that the user has access too.
        venv_kwargs["venv_bin"] = _win_user_where(
            account.username,
            account.password,
            "virtualenv",
        )
        venv_kwargs["python"] = _win_user_where(
            account.username,
            account.password,
            "python",
        )
    else:
        os.chown(str(tmp_path), account.info.uid, -1)

    # Create the virtual environment again as it should have been removed
    venv_create = create_virtualenv(
        str(venv_dir), user=account.username, password=account.password, **venv_kwargs
    )
    if venv_create.get("retcode", 1) > 0:
        pytest.skip(f"failed to create testcase virtual environment: {venv_create}")

    # pip install using a requirements file
    contents = "pep8\n"
    with pytest.helpers.temp_file("issue-6912-requirements.txt", contents, state_tree):
        ret = states.pip.installed(
            name="",
            user=account.username,
            password=account.password,
            bin_env=str(venv_dir),
            requirements="salt://issue-6912-requirements.txt",
        )
        assert ret.result is True

    for globmatch in (
        os.path.join(str(venv_dir), "**", "pep8*"),
        os.path.join(str(venv_dir), "*", "**", "pep8*"),
        os.path.join(str(venv_dir), "*", "*", "**", "pep8*"),
    ):
        for path in glob.glob(globmatch):
            if salt.utils.platform.is_windows():
                assert salt.utils.win_dacl.get_owner(path) == account.username
            else:
                assert os.stat(path).st_uid == account.info.uid


@pytest.mark.destructive_test
@pytest.mark.slow_test
def test_issue_6833_pip_upgrade_pip(tmp_path, create_virtualenv, modules, states):
    # Create the testing virtualenv
    if sys.platform == "win32":
        # To keeps the path short, we'll create this directory on the root
        # of the system drive. Otherwise the path is too long and the pip
        # upgrade will fail. Also, I don't know why salt.utils.platform
        # doesn't work in this function, that's why I used sys.platform
        # Need to use os.sep.join here instead of os.path.join because of
        # the colon in SystemDrive
        venv_dir = os.sep.join([os.environ["SystemDrive"], "tmp-6833-pip-upgrade-pip"])
    else:
        venv_dir = str(tmp_path / "6833-pip-upgrade-pip")
    ret = create_virtualenv(venv_dir)

    assert ret["retcode"] == 0

    if not (
        "New python executable" in ret["stdout"]
        or "created virtual environment" in ret["stdout"]
    ):
        assert (
            False
        ), "Expected STDOUT did not match. Full return dictionary:\n{}".format(
            pprint.pformat(ret)
        )

    # Let's install a fixed version pip over whatever pip was
    # previously installed
    ret = modules.pip.install("pip==19.3.1", upgrade=True, bin_env=venv_dir)

    if not isinstance(ret, dict):
        pytest.fail(
            "The 'pip.install' command did not return the excepted dictionary."
            " Output:\n{}".format(ret)
        )

    assert ret["retcode"] == 0
    assert "Successfully installed pip" in ret["stdout"]

    # Let's make sure we have pip 9.0.1 installed
    assert modules.pip.list("pip", bin_env=venv_dir) == {"pip": "19.3.1"}

    # Now the actual pip upgrade pip test
    ret = states.pip.installed(name="pip==20.0.1", upgrade=True, bin_env=venv_dir)

    if not isinstance(ret.raw, dict):
        pytest.fail(
            "The 'pip.install' command did not return the excepted dictionary."
            " Output:\n{}".format(ret)
        )

    assert ret.result is True
    assert ret.changes == {"pip==20.0.1": "Installed"}


@pytest.mark.slow_test
def test_pip_installed_specific_env(state_tree_prod, states, venv):
    contents = "pep8\n"

    # Let's write a requirements file
    with pytest.helpers.temp_file(
        "prod-env-requirements.txt", contents, state_tree_prod
    ):

        # The requirements file should not be found the base environment
        ret = states.pip.installed(
            name="",
            bin_env=str(venv.venv_dir),
            requirements="salt://prod-env-requirements.txt",
        )
        assert ret.result is False
        assert "'salt://prod-env-requirements.txt' not found" in ret.comment

        # The requirements file must be found in the prod environment
        ret = states.pip.installed(
            name="",
            bin_env=str(venv.venv_dir),
            saltenv="prod",
            requirements="salt://prod-env-requirements.txt",
        )
        assert ret.result is True
        assert (
            "Successfully processed requirements file salt://prod-env-requirements.txt"
            in ret.comment
        )

        # We're using the base environment but we're passing the prod
        # environment as an url arg to salt://
        ret = states.pip.installed(
            name="",
            bin_env=str(venv.venv_dir),
            requirements="salt://prod-env-requirements.txt?saltenv=prod",
        )
        assert ret.result is True
        assert "Requirements were already installed." in ret.comment


@pytest.mark.slow_test
def test_22359_pip_installed_unless_does_not_trigger_warnings(states, venv):
    # This test case should be moved to a format_call unit test specific to
    # the state internal keywords
    false_cmd = salt.utils.path.which("false")
    if salt.utils.platform.is_windows():
        false_cmd = "exit 1 >nul"
    ret = states.pip.installed(
        name="pep8",
        bin_env=str(venv.venv_dir),
        unless=false_cmd,
        timeout=600,
    )
    assert ret.result is True
    assert "warnings" not in next(iter(ret.raw.values()))


@pytest.mark.windows_whitelisted
@pytest.mark.slow_test
def test_issue_54755(tmp_path, state_tree, modules):
    """
    Verify github issue 54755 is resolved. This only fails when there is no
    pip module in the python environment. Since the test suite normally has
    a pip module this test will pass and is here for posterity. See also

    unit.states.test_pip_state.PipStateUtilsTest.test_pip_purge_method_with_pip

     and

    unit.states.test_pip_state.PipStateUtilsTest.test_pip_purge_method_without_pip

    Which also validate this issue and will pass/fail regardless of whether
    or not pip is installed.
    """
    file_path = tmp_path / "issue-54755"
    sls_contents = """issue-54755:
    file.managed:
        - name: {{ pillar['file_path'] }}
        - contents: issue-54755
        - unless: /bin/bash -c false
    """

    with pytest.helpers.temp_file("issue-54755.sls", sls_contents, state_tree):
        ret = modules.state.sls(mods="issue-54755", pillar={"file_path": file_path})
        key = f"file_|-issue-54755_|-{file_path}_|-managed"
        assert key in ret.raw
        assert ret.raw[key]["result"] is True
        with salt.utils.files.fopen(str(file_path), "r") as fp:
            assert fp.read().strip() == "issue-54755"



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>

     TestCase for salt.modules.dnsutil
"""

import pytest

import salt.modules.dnsutil as dnsutil
import salt.utils.stringutils
from tests.support.mock import MagicMock, mock_open, patch


@pytest.fixture
def mock_hosts_file():
    return (
        "##\n"
        "# Host Database\n"
        "#\n"
        "# localhost is used to configure the loopback interface\n"
        "# when the system is booting.  Do not change this entry.\n"
        "##\n"
        "127.0.0.1	localhost\n"
        "255.255.255.255	broadcasthost\n"
        "::1             localhost\n"
        "fe80::1%lo0	localhost"
    )


@pytest.fixture
def mock_hosts_file_rtn():
    return {
        "::1": ["localhost"],
        "255.255.255.255": ["broadcasthost"],
        "127.0.0.1": ["localhost"],
        "fe80::1%lo0": ["localhost"],
    }


@pytest.fixture
def mock_soa_zone():
    return (
        "$TTL 3D\n"
        "@               IN      SOA     land-5.com. root.land-5.com. (\n"
        "199609203       ; Serial\n"
        "28800   ; Refresh\n"
        "7200    ; Retry\n"
        "604800  ; Expire\n"
        "86400)  ; Minimum TTL\n"
        "NS      land-5.com.\n\n"
        "1                       PTR     localhost."
    )


@pytest.fixture
def mock_writes_list():
    return [
        "##\n",
        "# Host Database\n",
        "#\n",
        "# localhost is used to configure the loopback interface\n",
        "# when the system is booting.  Do not change this entry.\n",
        "##\n",
        "127.0.0.1 localhost",
        "\n",
        "255.255.255.255 broadcasthost",
        "\n",
        "::1 localhost",
        "\n",
        "fe80::1%lo0 localhost",
        "\n",
    ]


@pytest.fixture
def configure_loader_modules():
    return {dnsutil: {}}


def test_parse_hosts(mock_hosts_file):
    with patch("salt.utils.files.fopen", mock_open(read_data=mock_hosts_file)):
        assert dnsutil.parse_hosts() == {
            "::1": ["localhost"],
            "255.255.255.255": ["broadcasthost"],
            "127.0.0.1": ["localhost"],
            "fe80::1%lo0": ["localhost"],
        }


def test_hosts_append(mock_hosts_file, mock_hosts_file_rtn):
    with patch(
        "salt.utils.files.fopen", mock_open(read_data=mock_hosts_file)
    ) as m_open, patch(
        "salt.modules.dnsutil.parse_hosts",
        MagicMock(return_value=mock_hosts_file_rtn),
    ):
        dnsutil.hosts_append("/etc/hosts", "127.0.0.1", "ad1.yuk.co,ad2.yuk.co")
        writes = m_open.write_calls()
        # We should have called .write() only once, with the expected
        # content
        num_writes = len(writes)
        assert num_writes == 1, num_writes
        expected = salt.utils.stringutils.to_str("\n127.0.0.1 ad1.yuk.co ad2.yuk.co")
        assert writes[0] == expected, writes[0]


def test_hosts_remove(mock_hosts_file, mock_writes_list):
    to_remove = "ad1.yuk.co"
    new_mock_file = mock_hosts_file + "\n127.0.0.1 " + to_remove + "\n"
    with patch("salt.utils.files.fopen", mock_open(read_data=new_mock_file)) as m_open:
        dnsutil.hosts_remove("/etc/hosts", to_remove)
        writes = m_open.write_calls()
        assert writes == mock_writes_list, writes


def test_to_seconds_hour():
    assert dnsutil._to_seconds("4H") == 14400, "Did not detect valid hours as invalid"


def test_to_seconds_day():
    assert dnsutil._to_seconds("1D") == 86400, "Did not detect valid day as invalid"


def test_to_seconds_week():
    assert (
        dnsutil._to_seconds("2W") == 604800
    ), "Did not set time greater than one week to one week"


def test_to_seconds_empty():
    assert dnsutil._to_seconds("") == 604800, "Did not set empty time to one week"


def test_to_seconds_large():
    assert (
        dnsutil._to_seconds("604801") == 604800
    ), "Did not set time greater than one week to one week"



"""
Unit tests for salt.auth.ldap
"""

import pytest

import salt.auth.ldap
from tests.support.mock import patch

pytestmark = [
    pytest.mark.skipif(
        not salt.auth.ldap.HAS_LDAP, reason="Install python-ldap for this test"
    ),
]


class Bind:
    """
    fake search_s return
    """

    @staticmethod
    def search_s(*args, **kwargs):
        return [
            (
                "cn=saltusers,cn=groups,cn=compat,dc=saltstack,dc=com",
                {"memberUid": [b"saltuser"], "cn": [b"saltusers"]},
            ),
        ]


@pytest.fixture
def configure_loader_modules():
    return {
        salt.auth.ldap: {
            "__opts__": {
                "auth.ldap.binddn": (
                    "uid={{username}},cn=users,cn=compat,dc=saltstack,dc=com"
                ),
                "auth.ldap.port": 389,
                "auth.ldap.tls": False,
                "auth.ldap.server": "172.18.0.2",
                "auth.ldap.accountattributename": "memberUid",
                "auth.ldap.groupattribute": "memberOf",
                "auth.ldap.group_basedn": "cn=groups,cn=compat,dc=saltstack,dc=com",
                "auth.ldap.basedn": "dc=saltstack,dc=com",
                "auth.ldap.group_filter": (
                    "(&(memberUid={{ username }})(objectClass=posixgroup))"
                ),
            },
        },
    }


def test_config():
    """
    Test that the _config function works correctly
    """
    assert salt.auth.ldap._config("basedn") == "dc=saltstack,dc=com"
    assert (
        salt.auth.ldap._config("group_filter")
        == "(&(memberUid={{ username }})(objectClass=posixgroup))"
    )
    assert salt.auth.ldap._config("accountattributename") == "memberUid"
    assert salt.auth.ldap._config("groupattribute") == "memberOf"


def test_groups_freeipa():
    """
    test groups in freeipa
    """
    with patch.dict(salt.auth.ldap.__opts__, {"auth.ldap.freeipa": True}):
        with patch("salt.auth.ldap._bind", return_value=Bind):
            assert "saltusers" in salt.auth.ldap.groups("saltuser", password="password")


def test_groups():
    """
    test groups in ldap
    """
    with patch("salt.auth.ldap._bind", return_value=Bind):
        assert "saltusers" in salt.auth.ldap.groups("saltuser", password="password")


def test_groups_activedirectory():
    """
    test groups in activedirectory
    """
    with patch.dict(salt.auth.ldap.__opts__, {"auth.ldap.activedirectory": True}):
        with patch("salt.auth.ldap._bind", return_value=Bind):
            assert "saltusers" in salt.auth.ldap.groups("saltuser", password="password")


def test_auth_nopass():
    with patch.dict(salt.auth.ldap.__opts__, {"auth.ldap.bindpw": "p@ssw0rd!"}):
        with patch("salt.auth.ldap._bind_for_search", return_value=Bind):
            assert not salt.auth.ldap.auth("foo", None)


def test_auth_nouser():
    with patch.dict(salt.auth.ldap.__opts__, {"auth.ldap.bindpw": "p@ssw0rd!"}):
        with patch("salt.auth.ldap._bind_for_search", return_value=Bind):
            assert not salt.auth.ldap.auth(None, "foo")


def test_auth_nouserandpass():
    with patch.dict(salt.auth.ldap.__opts__, {"auth.ldap.bindpw": "p@ssw0rd!"}):
        with patch("salt.auth.ldap._bind_for_search", return_value=Bind):
            assert not salt.auth.ldap.auth(None, None)



"""
Functions for manipulating or otherwise processing strings
"""

import base64
import difflib
import errno
import fnmatch
import logging
import os
import re
import shlex
import time
import unicodedata

from salt.utils.decorators.jinja import jinja_filter

log = logging.getLogger(__name__)


@jinja_filter("to_bytes")
def to_bytes(s, encoding=None, errors="strict"):
    """
    Given bytes, bytearray, str, or unicode (python 2), return bytes (str for
    python 2)
    """
    if encoding is None:
        # Try utf-8 first, and fall back to detected encoding
        encoding = ("utf-8", __salt_system_encoding__)
    if not isinstance(encoding, (tuple, list)):
        encoding = (encoding,)

    if not encoding:
        raise ValueError("encoding cannot be empty")

    exc = None
    if isinstance(s, bytes):
        return s
    if isinstance(s, bytearray):
        return bytes(s)
    if isinstance(s, str):
        for enc in encoding:
            try:
                return s.encode(enc, errors)
            except UnicodeEncodeError as err:
                exc = err
                continue
        # The only way we get this far is if a UnicodeEncodeError was
        # raised, otherwise we would have already returned (or raised some
        # other exception).
        raise exc  # pylint: disable=raising-bad-type
    raise TypeError(f"expected str, bytes, or bytearray not {type(s)}")


def to_str(s, encoding=None, errors="strict", normalize=False):
    """
    Given str, bytes, bytearray, or unicode (py2), return str
    """

    def _normalize(s):
        try:
            return unicodedata.normalize("NFC", s) if normalize else s
        except TypeError:
            return s

    if encoding is None:
        # Try utf-8 first, and fall back to detected encoding
        encoding = ("utf-8", __salt_system_encoding__)
    if not isinstance(encoding, (tuple, list)):
        encoding = (encoding,)

    if not encoding:
        raise ValueError("encoding cannot be empty")

    if isinstance(s, str):
        return _normalize(s)

    exc = None
    if isinstance(s, (bytes, bytearray)):
        for enc in encoding:
            try:
                return _normalize(s.decode(enc, errors))
            except UnicodeDecodeError as err:
                exc = err
                continue
        # The only way we get this far is if a UnicodeDecodeError was
        # raised, otherwise we would have already returned (or raised some
        # other exception).
        raise exc  # pylint: disable=raising-bad-type
    raise TypeError(f"expected str, bytes, or bytearray not {type(s)}")


def to_unicode(s, encoding=None, errors="strict", normalize=False):
    """
    Given str or unicode, return unicode (str for python 3)
    """

    def _normalize(s):
        return unicodedata.normalize("NFC", s) if normalize else s

    if encoding is None:
        # Try utf-8 first, and fall back to detected encoding
        encoding = ("utf-8", __salt_system_encoding__)
    if not isinstance(encoding, (tuple, list)):
        encoding = (encoding,)

    if not encoding:
        raise ValueError("encoding cannot be empty")

    if isinstance(s, str):
        return _normalize(s)
    elif isinstance(s, (bytes, bytearray)):
        return _normalize(to_str(s, encoding, errors))
    raise TypeError(f"expected str, bytes, or bytearray not {type(s)}")


@jinja_filter("str_to_num")
@jinja_filter("to_num")
def to_num(text):
    """
    Convert a string to a number.
    Returns an integer if the string represents an integer, a floating
    point number if the string is a real number, or the string unchanged
    otherwise.
    """
    try:
        return int(text)
    except ValueError:
        try:
            return float(text)
        except ValueError:
            return text


def to_none(text):
    """
    Convert a string to None if the string is empty or contains only spaces.
    """
    if str(text).strip():
        return text
    return None


def is_quoted(value):
    """
    Return a single or double quote, if a string is wrapped in extra quotes.
    Otherwise return an empty string.
    """
    ret = ""
    if (
        isinstance(value, str)
        and value[0] == value[-1]
        and value.startswith(("'", '"'))
    ):
        ret = value[0]
    return ret


def dequote(value):
    """
    Remove extra quotes around a string.
    """
    if is_quoted(value):
        return value[1:-1]
    return value


@jinja_filter("is_hex")
def is_hex(value):
    """
    Returns True if value is a hexadecimal string, otherwise returns False
    """
    try:
        int(value, 16)
        return True
    except (TypeError, ValueError):
        return False


def is_binary(data):
    """
    Detects if the passed string of data is binary or text
    """
    if not data or not isinstance(data, ((str,), bytes)):
        return False

    if isinstance(data, bytes):
        if b"\0" in data:
            return True
    elif "\0" in data:
        return True

    text_characters = "".join([chr(x) for x in range(32, 127)] + list("\n\r\t\b"))
    # Get the non-text characters (map each character to itself then use the
    # 'remove' option to get rid of the text characters.)
    if isinstance(data, bytes):
        import salt.utils.data

        nontext = data.translate(None, salt.utils.data.encode(text_characters))
    else:
        trans = "".maketrans("", "", text_characters)
        nontext = data.translate(trans)

    # If more than 30% non-text characters, then
    # this is considered binary data
    if float(len(nontext)) / len(data) > 0.30:
        return True
    return False


@jinja_filter("random_str")
def random(size=32):
    key = os.urandom(size)
    return to_unicode(base64.b64encode(key).replace(b"\n", b"")[:size])


@jinja_filter("contains_whitespace")
def contains_whitespace(text):
    """
    Returns True if there are any whitespace characters in the string
    """
    return any(x.isspace() for x in text)


@jinja_filter("human_to_bytes")
def human_to_bytes(size, default_unit="B", handle_metric=False):
    """
    Given a human-readable byte string (e.g. 2G, 30MB, 64KiB),
    return the number of bytes.  Will return 0 if the argument has
    unexpected form.

    .. versionadded:: 2018.3.0
    .. versionchanged:: 3005
    """
    m = re.match(r"(?P<value>[0-9.]*)\s*(?P<unit>.*)$", str(size).strip())
    value = m.group("value")
    # default unit
    unit = m.group("unit").lower() or default_unit.lower()
    try:
        value = int(value)
    except ValueError:
        try:
            value = float(value)
        except ValueError:
            return 0
    # flag for base ten
    dec = False
    if re.match(r"[kmgtpezy]b$", unit):
        dec = True if handle_metric else False
    elif not re.match(r"(b|[kmgtpezy](ib)?)$", unit):
        return 0
    p = "bkmgtpezy".index(unit[0])
    value *= 10 ** (p * 3) if dec else 2 ** (p * 10)
    return int(value)


def build_whitespace_split_regex(text):
    '''
    Create a regular expression at runtime which should match ignoring the
    addition or deletion of white space or line breaks, unless between commas

    Example:

    .. code-block:: python

        >>> import re
        >>> import salt.utils.stringutils
        >>> regex = salt.utils.stringutils.build_whitespace_split_regex(
        ...     """if [ -z "$debian_chroot" ] && [ -r /etc/debian_chroot ]; then"""
        ... )

        >>> regex
        '(?:[\\s]+)?if(?:[\\s]+)?\\[(?:[\\s]+)?\\-z(?:[\\s]+)?\\"\\$debian'
        '\\_chroot\\"(?:[\\s]+)?\\](?:[\\s]+)?\\&\\&(?:[\\s]+)?\\[(?:[\\s]+)?'
        '\\-r(?:[\\s]+)?\\/etc\\/debian\\_chroot(?:[\\s]+)?\\]\\;(?:[\\s]+)?'
        'then(?:[\\s]+)?'
        >>> re.search(
        ...     regex,
        ...     """if [ -z "$debian_chroot" ] && [ -r /etc/debian_chroot ]; then"""
        ... )

        <_sre.SRE_Match object at 0xb70639c0>
        >>>

    '''

    def __build_parts(text):
        lexer = shlex.shlex(text)
        lexer.whitespace_split = True
        lexer.commenters = ""
        if r"'\"" in text:
            lexer.quotes = ""
        elif "'" in text:
            lexer.quotes = '"'
        elif '"' in text:
            lexer.quotes = "'"
        return list(lexer)

    regex = r""
    for line in text.splitlines():
        parts = [re.escape(s) for s in __build_parts(line)]
        regex += r"(?:[\s]+)?{}(?:[\s]+)?".format(r"(?:[\s]+)?".join(parts))
    return rf"(?m)^{regex}$"


def expr_match(line, expr):
    """
    Checks whether or not the passed value matches the specified expression.
    Tries to match expr first as a glob using fnmatch.fnmatch(), and then tries
    to match expr as a regular expression. Originally designed to match minion
    IDs for whitelists/blacklists.

    Note that this also does exact matches, as fnmatch.fnmatch() will return
    ``True`` when no glob characters are used and the string is an exact match:

    .. code-block:: python

        >>> fnmatch.fnmatch('foo', 'foo')
        True
    """
    try:
        if fnmatch.fnmatch(line, expr):
            return True
        try:
            if re.match(rf"\A{expr}\Z", line):
                return True
        except re.error:
            pass
    except TypeError:
        log.exception("Value %r or expression %r is not a string", line, expr)
    return False


@jinja_filter("check_whitelist_blacklist")
def check_whitelist_blacklist(value, whitelist=None, blacklist=None):
    """
    Check a whitelist and/or blacklist to see if the value matches it.

    value
        The item to check the whitelist and/or blacklist against.

    whitelist
        The list of items that are white-listed. If ``value`` is found
        in the whitelist, then the function returns ``True``. Otherwise,
        it returns ``False``.

    blacklist
        The list of items that are black-listed. If ``value`` is found
        in the blacklist, then the function returns ``False``. Otherwise,
        it returns ``True``.

    If both a whitelist and a blacklist are provided, value membership
    in the blacklist will be examined first. If the value is not found
    in the blacklist, then the whitelist is checked. If the value isn't
    found in the whitelist, the function returns ``False``.
    """
    # Normalize the input so that we have a list
    if blacklist:
        if isinstance(blacklist, str):
            blacklist = [blacklist]
        if not hasattr(blacklist, "__iter__"):
            raise TypeError(
                "Expecting iterable blacklist, but got {} ({})".format(
                    type(blacklist).__name__, blacklist
                )
            )
    else:
        blacklist = []

    if whitelist:
        if isinstance(whitelist, str):
            whitelist = [whitelist]
        if not hasattr(whitelist, "__iter__"):
            raise TypeError(
                "Expecting iterable whitelist, but got {} ({})".format(
                    type(whitelist).__name__, whitelist
                )
            )
    else:
        whitelist = []

    _blacklist_match = any(expr_match(value, expr) for expr in blacklist)
    _whitelist_match = any(expr_match(value, expr) for expr in whitelist)

    if blacklist and not whitelist:
        # Blacklist but no whitelist
        return not _blacklist_match
    elif whitelist and not blacklist:
        # Whitelist but no blacklist
        return _whitelist_match
    elif blacklist and whitelist:
        # Both whitelist and blacklist
        return not _blacklist_match and _whitelist_match
    else:
        # No blacklist or whitelist passed
        return True


def check_include_exclude(path_str, include_pat=None, exclude_pat=None):
    """
    Check for glob or regexp patterns for include_pat and exclude_pat in the
    'path_str' string and return True/False conditions as follows.
      - Default: return 'True' if no include_pat or exclude_pat patterns are
        supplied
      - If only include_pat or exclude_pat is supplied: return 'True' if string
        passes the include_pat test or fails exclude_pat test respectively
      - If both include_pat and exclude_pat are supplied: return 'True' if
        include_pat matches AND exclude_pat does not match
    """

    def _pat_check(path_str, check_pat):
        if re.match("E@", check_pat):
            return True if re.search(check_pat[2:], path_str) else False
        else:
            return True if fnmatch.fnmatch(path_str, check_pat) else False

    ret = True  # -- default true
    # Before pattern match, check if it is regexp (E@'') or glob(default)
    if include_pat:
        if isinstance(include_pat, list):
            for include_line in include_pat:
                retchk_include = _pat_check(path_str, include_line)
                if retchk_include:
                    break
        else:
            retchk_include = _pat_check(path_str, include_pat)

    if exclude_pat:
        if isinstance(exclude_pat, list):
            for exclude_line in exclude_pat:
                retchk_exclude = not _pat_check(path_str, exclude_line)
                if not retchk_exclude:
                    break
        else:
            retchk_exclude = not _pat_check(path_str, exclude_pat)

    # Now apply include/exclude conditions
    if include_pat and not exclude_pat:
        ret = retchk_include
    elif exclude_pat and not include_pat:
        ret = retchk_exclude
    elif include_pat and exclude_pat:
        ret = retchk_include and retchk_exclude
    else:
        ret = True

    return ret


def print_cli(msg, retries=10, step=0.01):
    """
    Wrapper around print() that suppresses tracebacks on broken pipes (i.e.
    when salt output is piped to less and less is stopped prematurely).
    """
    while retries:
        try:
            try:
                print(msg)
            except UnicodeEncodeError:
                print(msg.encode("utf-8"))
        except OSError as exc:
            err = f"{exc}"
            if exc.errno != errno.EPIPE:
                if (
                    "temporarily unavailable" in err or exc.errno in (errno.EAGAIN,)
                ) and retries:
                    time.sleep(step)
                    retries -= 1
                    continue
                else:
                    raise
        break


def get_context(template, line, num_lines=5, marker=None):
    """
    Returns debugging context around a line in a given string

    Returns:: string
    """
    template_lines = template.splitlines()
    num_template_lines = len(template_lines)

    # In test mode, a single line template would return a crazy line number like,
    # 357. Do this sanity check and if the given line is obviously wrong, just
    # return the entire template
    if line > num_template_lines:
        return template

    context_start = max(0, line - num_lines - 1)  # subt 1 for 0-based indexing
    context_end = min(num_template_lines, line + num_lines)
    error_line_in_context = line - context_start - 1  # subtr 1 for 0-based idx

    buf = []
    if context_start > 0:
        buf.append("[...]")
        error_line_in_context += 1

    buf.extend(template_lines[context_start:context_end])

    if context_end < num_template_lines:
        buf.append("[...]")

    if marker:
        buf[error_line_in_context] += marker

    return "---\n{}\n---".format("\n".join(buf))


def get_diff(a, b, *args, **kwargs):
    """
    Perform diff on two iterables containing lines from two files, and return
    the diff as as string. Lines are normalized to str types to avoid issues
    with unicode on PY2.
    """
    encoding = ("utf-8", "latin-1", __salt_system_encoding__)
    # Late import to avoid circular import
    import salt.utils.data

    return "".join(
        difflib.unified_diff(
            salt.utils.data.decode_list(a, encoding=encoding),
            salt.utils.data.decode_list(b, encoding=encoding),
            *args,
            **kwargs,
        )
    )


@jinja_filter("to_snake_case")
def camel_to_snake_case(camel_input):
    """
    Converts camelCase (or CamelCase) to snake_case.
    From https://codereview.stackexchange.com/questions/185966/functions-to-convert-camelcase-strings-to-snake-case

    :param str camel_input: The camelcase or CamelCase string to convert to snake_case

    :return str
    """
    res = camel_input[0].lower()
    for i, letter in enumerate(camel_input[1:], 1):
        if letter.isupper():
            if camel_input[i - 1].islower() or (
                i != len(camel_input) - 1 and camel_input[i + 1].islower()
            ):
                res += "_"
        res += letter.lower()
    return res


@jinja_filter("to_camelcase")
def snake_to_camel_case(snake_input, uppercamel=False):
    """
    Converts snake_case to camelCase (or CamelCase if uppercamel is ``True``).
    Inspired by https://codereview.stackexchange.com/questions/85311/transform-snake-case-to-camelcase

    :param str snake_input: The input snake_case string to convert to camelCase
    :param bool uppercamel: Whether or not to convert to CamelCase instead

    :return str
    """
    words = snake_input.split("_")
    if uppercamel:
        words[0] = words[0].capitalize()
    return words[0] + "".join(word.capitalize() for word in words[1:])



"""
    tests.pytests.unit.beacons.test_adb
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    ADB beacon test cases
"""

import pytest

import salt.beacons.adb as adb
from tests.support.mock import Mock, patch


@pytest.fixture
def configure_loader_modules():
    return {adb: {"last_state": {}, "last_state_extra": {"no_devices": False}}}


def test_no_adb_command():
    with patch("salt.utils.path.which") as mock:
        mock.return_value = None

        ret = adb.__virtual__()

        mock.assert_called_once_with("adb")
        assert ret == (False, "adb is missing.")


def test_with_adb_command():
    with patch("salt.utils.path.which") as mock:
        mock.return_value = "/usr/bin/adb"

        ret = adb.__virtual__()

        mock.assert_called_once_with("adb")
        assert ret == "adb"


def test_non_list_config():
    config = {}

    ret = adb.validate(config)
    assert ret == (False, "Configuration for adb beacon must be a list.")


def test_empty_config():
    config = [{}]

    ret = adb.validate(config)
    assert ret == (False, "Configuration for adb beacon must include a states array.")


def test_invalid_states():
    config = [{"states": ["Random", "Failings"]}]

    ret = adb.validate(config)
    assert ret == (
        False,
        "Need a one of the following adb states: offline, bootloader, device, host,"
        " recovery, no permissions, sideload, unauthorized, unknown, missing",
    )


def test_device_state():
    config = [{"states": ["device"]}]

    mock = Mock(return_value="List of devices attached\nHTC\tdevice")
    with patch.dict(adb.__salt__, {"cmd.run": mock}):
        ret = adb.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = adb.beacon(config)
        assert ret == [{"device": "HTC", "state": "device", "tag": "device"}]


def test_device_state_change():
    config = [{"states": ["offline"]}]

    out = [
        "List of devices attached\nHTC\tdevice",
        "List of devices attached\nHTC\toffline",
    ]

    mock = Mock(side_effect=out)
    with patch.dict(adb.__salt__, {"cmd.run": mock}):
        ret = adb.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = adb.beacon(config)
        assert ret == []

        ret = adb.beacon(config)
        assert ret == [{"device": "HTC", "state": "offline", "tag": "offline"}]


def test_multiple_devices():
    config = [{"states": ["offline", "device"]}]

    out = [
        "List of devices attached\nHTC\tdevice",
        "List of devices attached\nHTC\toffline\nNexus\tdevice",
    ]

    mock = Mock(side_effect=out)
    with patch.dict(adb.__salt__, {"cmd.run": mock}):
        ret = adb.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = adb.beacon(config)
        assert ret == [{"device": "HTC", "state": "device", "tag": "device"}]

        ret = adb.beacon(config)
        assert ret == [
            {"device": "HTC", "state": "offline", "tag": "offline"},
            {"device": "Nexus", "state": "device", "tag": "device"},
        ]


def test_no_devices_with_different_states():
    config = [{"states": ["offline"], "no_devices_event": True}]

    mock = Mock(return_value="List of devices attached\nHTC\tdevice")
    with patch.dict(adb.__salt__, {"cmd.run": mock}):
        ret = adb.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = adb.beacon(config)
        assert ret == []


def test_no_devices_no_repeat():
    config = [{"states": ["offline", "device"], "no_devices_event": True}]

    out = [
        "List of devices attached\nHTC\tdevice",
        "List of devices attached",
        "List of devices attached",
    ]

    mock = Mock(side_effect=out)
    with patch.dict(adb.__salt__, {"cmd.run": mock}):
        ret = adb.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = adb.beacon(config)
        assert ret == [{"device": "HTC", "state": "device", "tag": "device"}]

        ret = adb.beacon(config)
        assert ret == [{"tag": "no_devices"}]

        ret = adb.beacon(config)
        assert ret == []


def test_no_devices():
    config = [{"states": ["offline", "device"], "no_devices_event": True}]

    out = ["List of devices attached", "List of devices attached"]

    mock = Mock(side_effect=out)
    with patch.dict(adb.__salt__, {"cmd.run": mock}):
        ret = adb.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = adb.beacon(config)
        assert ret == [{"tag": "no_devices"}]

        ret = adb.beacon(config)
        assert ret == []


def test_device_missing():
    config = [{"states": ["device", "missing"]}]

    out = [
        "List of devices attached\nHTC\tdevice",
        "List of devices attached",
        "List of devices attached\nHTC\tdevice",
        "List of devices attached\nHTC\tdevice",
    ]

    mock = Mock(side_effect=out)
    with patch.dict(adb.__salt__, {"cmd.run": mock}):
        ret = adb.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = adb.beacon(config)
        assert ret == [{"device": "HTC", "state": "device", "tag": "device"}]

        ret = adb.beacon(config)
        assert ret == [{"device": "HTC", "state": "missing", "tag": "missing"}]

        ret = adb.beacon(config)
        assert ret == [{"device": "HTC", "state": "device", "tag": "device"}]

        ret = adb.beacon(config)
        assert ret == []


def test_with_startup():
    config = [{"states": ["device"]}]

    mock = Mock(
        return_value=(
            "* daemon started successfully *\nList of devices attached\nHTC\tdevice"
        ),
    )
    with patch.dict(adb.__salt__, {"cmd.run": mock}):
        ret = adb.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = adb.beacon(config)
        assert ret == [{"device": "HTC", "state": "device", "tag": "device"}]


def test_with_user():
    config = [{"states": ["device"], "user": "fred"}]

    mock = Mock(
        return_value=(
            "* daemon started successfully *\nList of devices attached\nHTC\tdevice"
        )
    )
    with patch.dict(adb.__salt__, {"cmd.run": mock}):
        ret = adb.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = adb.beacon(config)
        mock.assert_called_once_with("adb devices", runas="fred")
        assert ret == [{"device": "HTC", "state": "device", "tag": "device"}]


def test_device_low_battery():
    config = [{"states": ["device"], "battery_low": 30}]

    out = [
        "List of devices attached\nHTC\tdevice",
        "25",
    ]

    mock = Mock(side_effect=out)
    with patch.dict(adb.__salt__, {"cmd.run": mock}):
        ret = adb.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = adb.beacon(config)
        assert ret == [
            {"device": "HTC", "state": "device", "tag": "device"},
            {"device": "HTC", "battery_level": 25, "tag": "battery_low"},
        ]


def test_device_no_repeat():
    config = [{"states": ["device"], "battery_low": 30}]

    out = [
        "List of devices attached\nHTC\tdevice",
        "25",
        "List of devices attached\nHTC\tdevice",
        "25",
    ]

    mock = Mock(side_effect=out)
    with patch.dict(adb.__salt__, {"cmd.run": mock}):
        ret = adb.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = adb.beacon(config)
        assert ret == [
            {"device": "HTC", "state": "device", "tag": "device"},
            {"device": "HTC", "battery_level": 25, "tag": "battery_low"},
        ]

        ret = adb.beacon(config)
        assert ret == []


def test_device_no_repeat_capacity_increase():
    config = [{"states": ["device"], "battery_low": 75}]

    out = [
        "List of devices attached\nHTC\tdevice",
        "25",
        "List of devices attached\nHTC\tdevice",
        "30",
    ]

    mock = Mock(side_effect=out)
    with patch.dict(adb.__salt__, {"cmd.run": mock}):
        ret = adb.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = adb.beacon(config)
        assert ret == [
            {"device": "HTC", "state": "device", "tag": "device"},
            {"device": "HTC", "battery_level": 25, "tag": "battery_low"},
        ]

        ret = adb.beacon(config)
        assert ret == []


def test_device_no_repeat_with_not_found_state():
    config = [{"states": ["offline"], "battery_low": 30}]

    out = [
        "List of devices attached\nHTC\tdevice",
        "25",
        "List of devices attached\nHTC\tdevice",
        "25",
    ]

    mock = Mock(side_effect=out)
    with patch.dict(adb.__salt__, {"cmd.run": mock}):
        ret = adb.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = adb.beacon(config)
        assert ret == [{"device": "HTC", "battery_level": 25, "tag": "battery_low"}]

        ret = adb.beacon(config)
        assert ret == []


def test_device_battery_charged():
    config = [{"states": ["device"], "battery_low": 30}]

    out = [
        "List of devices attached\nHTC\tdevice",
        "100",
    ]

    mock = Mock(side_effect=out)
    with patch.dict(adb.__salt__, {"cmd.run": mock}):
        ret = adb.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = adb.beacon(config)
        assert ret == [{"device": "HTC", "state": "device", "tag": "device"}]


def test_device_low_battery_equal():
    config = [{"states": ["device"], "battery_low": 25}]

    out = [
        "List of devices attached\nHTC\tdevice",
        "25",
    ]

    mock = Mock(side_effect=out)
    with patch.dict(adb.__salt__, {"cmd.run": mock}):
        ret = adb.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = adb.beacon(config)
        assert ret == [
            {"device": "HTC", "state": "device", "tag": "device"},
            {"device": "HTC", "battery_level": 25, "tag": "battery_low"},
        ]


def test_device_battery_not_found():
    config = [{"states": ["device"], "battery_low": 25}]

    out = [
        "List of devices attached\nHTC\tdevice",
        "/system/bin/sh: cat: /sys/class/power_supply/*/capacity: No such file or"
        " directory",
    ]

    mock = Mock(side_effect=out)
    with patch.dict(adb.__salt__, {"cmd.run": mock}):
        ret = adb.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = adb.beacon(config)
        assert ret == [{"device": "HTC", "state": "device", "tag": "device"}]


def test_device_repeat_multi():
    config = [{"states": ["offline"], "battery_low": 35}]

    out = [
        "List of devices attached\nHTC\tdevice",
        "25",
        "List of devices attached\nHTC\tdevice",
        "40",
        "List of devices attached\nHTC\tdevice",
        "25",
        "List of devices attached\nHTC\tdevice",
        "80",
    ]

    mock = Mock(side_effect=out)
    with patch.dict(adb.__salt__, {"cmd.run": mock}):
        ret = adb.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = adb.beacon(config)
        assert ret == [{"device": "HTC", "battery_level": 25, "tag": "battery_low"}]

        ret = adb.beacon(config)
        assert ret == []

        ret = adb.beacon(config)
        assert ret == [{"device": "HTC", "battery_level": 25, "tag": "battery_low"}]

        ret = adb.beacon(config)
        assert ret == []


def test_weird_batteries():
    config = [{"states": ["device"], "battery_low": 25}]

    out = [
        "List of devices attached\nHTC\tdevice",
        "-9000",
    ]

    mock = Mock(side_effect=out)
    with patch.dict(adb.__salt__, {"cmd.run": mock}):
        ret = adb.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = adb.beacon(config)
        assert ret == [{"device": "HTC", "state": "device", "tag": "device"}]


def test_multiple_batteries():
    config = [{"states": ["device"], "battery_low": 30}]

    out = [
        "List of devices attached\nHTC\tdevice",
        "25\n40",
    ]

    mock = Mock(side_effect=out)
    with patch.dict(adb.__salt__, {"cmd.run": mock}):
        ret = adb.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = adb.beacon(config)
        assert ret == [
            {"device": "HTC", "state": "device", "tag": "device"},
            {"device": "HTC", "battery_level": 25, "tag": "battery_low"},
        ]


def test_multiple_low_batteries():
    config = [{"states": ["device"], "battery_low": 30}]

    out = [
        "List of devices attached\nHTC\tdevice",
        "25\n14",
    ]

    mock = Mock(side_effect=out)
    with patch.dict(adb.__salt__, {"cmd.run": mock}):
        ret = adb.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = adb.beacon(config)
        assert ret == [
            {"device": "HTC", "state": "device", "tag": "device"},
            {"device": "HTC", "battery_level": 25, "tag": "battery_low"},
        ]
